history = {'events': [
    {'eventId': 1,
     'eventTimestamp': 1419350960.696,
     'eventType': 'WorkflowExecutionStarted',
     'workflowExecutionStartedEventAttributes': {
         'childPolicy': 'TERMINATE',
         'executionStartToCloseTimeout': '3600',
         'parentInitiatedEventId': 0,
         'taskList': {'name': 'garcon_decider'},
         'taskStartToCloseTimeout': '300',
         'workflowType': {'name': 'garcon_decider', 'version': '1.0'}}},
    {'decisionTaskScheduledEventAttributes': {
        'startToCloseTimeout': '300',
        'taskList': {'name': 'garcon_decider'}},
     'eventId': 2,
     'eventTimestamp': 1419350960.696,
     'eventType': 'DecisionTaskScheduled'},
    {'decisionTaskStartedEventAttributes': {
        'scheduledEventId': 2},
     'eventId': 3,
     'eventTimestamp': 1419350960.767,
     'eventType': 'DecisionTaskStarted'},
    {'decisionTaskCompletedEventAttributes': {
        'scheduledEventId': 2, 'startedEventId': 3},
     'eventId': 4,
     'eventTimestamp': 1419350960.945,
     'eventType': 'DecisionTaskCompleted'},
    {'activityTaskScheduledEventAttributes': {
        'activityId': 'workflow_name_activity_1-1',
        'activityType': {'name': 'workflow_name_activity_1', 'version': '1.0'},
        'decisionTaskCompletedEventId': 4,
        'heartbeatTimeout': '600',
        'input': 'json object',
        'scheduleToCloseTimeout': '3900',
        'scheduleToStartTimeout': '300',
        'startToCloseTimeout': '3600',
        'taskList': {'name': 'garcon_workflow_name_activity_1'}},
     'eventId': 5,
     'eventTimestamp': 1419350960.945,
     'eventType': 'ActivityTaskScheduled'},
    {'activityTaskStartedEventAttributes': {
        'scheduledEventId': 5},
     'eventId': 6,
     'eventTimestamp': 1419351025.473,
     'eventType': 'ActivityTaskStarted'},
    {'activityTaskCompletedEventAttributes': {
        'scheduledEventId': 5, 'startedEventId': 6},
     'eventId': 7,
     'eventTimestamp': 1419351025.534,
     'eventType': 'ActivityTaskCompleted'},
    {'decisionTaskScheduledEventAttributes': {
        'startToCloseTimeout': '300',
        'taskList': {'name': 'garcon_decider'}},
     'eventId': 8,
     'eventTimestamp': 1419351025.534,
     'eventType': 'DecisionTaskScheduled'},
    {'decisionTaskStartedEventAttributes': {'scheduledEventId': 8},
     'eventId': 9,
     'eventTimestamp': 1419351026.066,
     'eventType': 'DecisionTaskStarted'},
    {'decisionTaskCompletedEventAttributes': {
        'scheduledEventId': 8, 'startedEventId': 9},
     'eventId': 10,
     'eventTimestamp': 1419351026.232,
     'eventType': 'DecisionTaskCompleted'},
    {'activityTaskScheduledEventAttributes': {
        'activityId': 'workflow_name_activity_2-1',
        'activityType': {'name': 'workflow_name_activity_2', 'version': '1.0'},
        'decisionTaskCompletedEventId': 10,
        'heartbeatTimeout': '600',
        'input': 'json object',
        'scheduleToCloseTimeout': '3900',
        'scheduleToStartTimeout': '300',
        'startToCloseTimeout': '3600',
        'taskList': {'name': 'garcon_workflow_name_activity_2'}},
     'eventId': 11,
     'eventTimestamp': 1419351026.232,
     'eventType': 'ActivityTaskScheduled'},
    {'activityTaskScheduledEventAttributes': {
        'activityId': 'workflow_name_activity_3-1',
        'activityType': {'name': 'workflow_name_activity_3', 'version': '1.0'},
        'decisionTaskCompletedEventId': 10,
        'heartbeatTimeout': '600',
        'input': 'json object',
        'scheduleToCloseTimeout': '3900',
        'scheduleToStartTimeout': '300',
        'startToCloseTimeout': '3600',
        'taskList': {'name': 'garcon_workflow_name_activity_3'}},
     'eventId': 12,
     'eventTimestamp': 1419351026.232,
     'eventType': 'ActivityTaskScheduled'},
    {'activityTaskStartedEventAttributes': {
        'scheduledEventId': 12},
     'eventId': 13,
     'eventTimestamp': 1419351026.357,
     'eventType': 'ActivityTaskStarted'},
    {'activityTaskCompletedEventAttributes': {
        'scheduledEventId': 12, 'startedEventId': 13},
     'eventId': 14,
     'eventTimestamp': 1419351026.437,
     'eventType': 'ActivityTaskCompleted'},
    {'decisionTaskScheduledEventAttributes': {
        'startToCloseTimeout': '300',
        'taskList': {'name': 'garcon_decider'}},
     'eventId': 15,
     'eventTimestamp': 1419351026.437,
     'eventType': 'DecisionTaskScheduled'},
    {'activityTaskStartedEventAttributes': {
        'scheduledEventId': 11},
     'eventId': 16,
     'eventTimestamp': 1419351026.542,
     'eventType': 'ActivityTaskStarted'},
    {'decisionTaskStartedEventAttributes': {
        'scheduledEventId': 15},
     'eventId': 17,
     'eventTimestamp': 1419351026.622,
     'eventType': 'DecisionTaskStarted'},
    {'decisionTaskCompletedEventAttributes': {
        'scheduledEventId': 15, 'startedEventId': 17},
     'eventId': 18,
     'eventTimestamp': 1419351026.803,
     'eventType': 'DecisionTaskCompleted'},
    {'activityTaskCompletedEventAttributes': {
        'scheduledEventId': 11, 'startedEventId': 16, 'result': '{"k": "v"}'},
     'eventId': 19,
     'eventTimestamp': 1419351026.932,
     'eventType': 'ActivityTaskCompleted'},
    {'decisionTaskScheduledEventAttributes': {
        'startToCloseTimeout': '300',
        'taskList': {'name': 'garcon_decider'}},
     'eventId': 20,
     'eventTimestamp': 1419351026.932,
     'eventType': 'DecisionTaskScheduled'},
    {'decisionTaskStartedEventAttributes': {'scheduledEventId': 20},
     'eventId': 21,
     'eventTimestamp': 1419351027.276,
     'eventType': 'DecisionTaskStarted'},
    {'decisionTaskCompletedEventAttributes': {
        'scheduledEventId': 20, 'startedEventId': 21},
     'eventId': 22,
     'eventTimestamp': 1419351027.381,
     'eventType': 'DecisionTaskCompleted'},
    {'activityTaskScheduledEventAttributes': {
        'activityId': 'workflow_name_activity_4-1',
        'activityType': {'name': 'workflow_name_activity_4', 'version': '1.0'},
        'decisionTaskCompletedEventId': 22,
        'heartbeatTimeout': '600',
        'input': 'json object',
        'scheduleToCloseTimeout': '3900',
        'scheduleToStartTimeout': '300',
        'startToCloseTimeout': '3600',
        'taskList': {'name': 'garcon_workflow_name_activity_4'}},
     'eventId': 23,
     'eventTimestamp': 1419351027.381,
     'eventType': 'ActivityTaskScheduled'},
    {'activityTaskStartedEventAttributes': {'scheduledEventId': 23},
     'eventId': 24,
     'eventTimestamp': 1419351027.43,
     'eventType': 'ActivityTaskStarted'},
    {'activityTaskCompletedEventAttributes': {
        'scheduledEventId': 23, 'startedEventId': 24},
     'eventId': 25,
     'eventTimestamp': 1419351027.492,
     'eventType': 'ActivityTaskCompleted'},
    {'decisionTaskScheduledEventAttributes': {
        'startToCloseTimeout': '300',
        'taskList': {'name': 'garcon_decider'}},
     'eventId': 26,
     'eventTimestamp': 1419351027.492,
     'eventType': 'DecisionTaskScheduled'},
    {'decisionTaskStartedEventAttributes': {'scheduledEventId': 26},
     'eventId': 27,
     'eventTimestamp': 1419351027.555,
     'eventType': 'DecisionTaskStarted'}],
    'workflowExecution': {
    'runId': '123abc=',
    'workflowId': 'test-workflow-id'}
}



try:
    from unittest.mock import MagicMock
except:
    from mock import MagicMock

import pytest
import boto3


@pytest.fixture
def boto_client(monkeypatch):
    """Create a fake boto client."""
    return MagicMock(spec=boto3.client('swf', region_name='us-east-1'))



import logging

from garcon import log
from tests.fixtures import log as fixture


def test_get_logger_namespace():
    """Test that the logger nsmaespace is generated properly for a given
    execution context.
    """

    assert log.get_logger_namespace(fixture.execution_context) == '.'.join([
        log.LOGGER_PREFIX,
        fixture.execution_context.get('execution.domain'),
        fixture.execution_context.get('execution.workflow_id'),
        fixture.execution_context.get('execution.run_id')])


def test_set_log_context():
    """Test that the logger_name property has set properly or has not been set
    if an invalid execution context is passed in.
    """

    valid_mock = fixture.log_enabled_object()
    invalid_mock = fixture.log_disabled_object()

    assert valid_mock.logger_name == '.'.join([
        log.LOGGER_PREFIX,
        fixture.execution_context.get('execution.domain'),
        fixture.execution_context.get('execution.workflow_id'),
        fixture.execution_context.get('execution.run_id')])
    assert valid_mock.logger is logging.getLogger(valid_mock.logger_name)

    assert getattr(invalid_mock, 'logger_name', None) is None
    assert invalid_mock.logger is logging.getLogger(log.LOGGER_PREFIX)


def test_unset_log_context():
    """Test that the logger_name property has been unset.
    """

    valid_mock = fixture.log_enabled_object()
    valid_mock.unset_log_context()

    assert getattr(valid_mock, 'logger_name', None) is None



# -*- coding: utf-8 -*-
from garcon import activity
from garcon import context
import json


def activity_states_from_events(events):
    """Get activity states from a list of events.

    The workflow events contains the different states of our activities. This
    method consumes the logs, and regenerates a dictionnary with the list of
    all the activities and their states.

    Note:
        Please note: from the list of events, only activities that have been
        registered are accessible. For all the others that have not yet
        started, they won't be part of this list.

    Args:
        events (dict): list of all the events.
    Return:
        `dict`: the activities and their state.
    """

    events = sorted(events, key=lambda item: item.get('eventId'))
    event_id_info = dict()
    activity_events = dict()

    for event in events:
        event_id = event.get('eventId')
        event_type = event.get('eventType')

        if event_type == 'ActivityTaskScheduled':
            activity_info = event.get('activityTaskScheduledEventAttributes')
            activity_id = activity_info.get('activityId')
            activity_name = activity_info.get('activityType').get('name')
            event_id_info.update({
                event_id: {
                    'activity_name': activity_name,
                    'activity_id': activity_id}
            })

            activity_events.setdefault(
                activity_name, {}).setdefault(
                    activity_id,
                    activity.ActivityState(activity_id)).add_state(
                        activity.ACTIVITY_SCHEDULED)

        elif event_type == 'ActivityTaskFailed':
            activity_info = event.get('activityTaskFailedEventAttributes')
            activity_event = event_id_info.get(
                activity_info.get('scheduledEventId'))
            activity_id = activity_event.get('activity_id')

            activity_events.setdefault(
                activity_event.get('activity_name'), {}).setdefault(
                    activity_id,
                    activity.ActivityState(activity_id)).add_state(
                        activity.ACTIVITY_FAILED)

        elif event_type == 'ActivityTaskCompleted':
            activity_info = event.get('activityTaskCompletedEventAttributes')
            activity_event = event_id_info.get(
                activity_info.get('scheduledEventId'))
            activity_id = activity_event.get('activity_id')

            activity_events.setdefault(
                activity_event.get('activity_name'), {}).setdefault(
                    activity_id,
                    activity.ActivityState(activity_id)).add_state(
                        activity.ACTIVITY_COMPLETED)

            result = json.loads(activity_info.get('result') or '{}')
            activity_events.get(
                activity_event.get('activity_name')).get(
                    activity_id).set_result(result)

    return activity_events


def get_current_context(events):
    """Get the current context from the list of events.

    Each activity returns bits of information that needs to be provided to the
    next activities.

    Args:
        events (list): List of events.
    Return:
        dict: The current context.
    """

    events = sorted(events, key=lambda item: item.get('eventId'))
    execution_context = context.ExecutionContext(events)
    return execution_context



"""
Param
=====

Params are values that are passed to the activities. They are either provided
by the execution context (see Param) or are statically provided at the runtime
of the activity (see StaticParam). Custom params should extend the Param class.

Note:
    Make sure the custom param class lists out all the dependencies to the
    execution context in `requirements`.
"""


class BaseParam:
    """Base Param Class.

    Provides the structure and required methods of any param class.
    """

    @property
    def requirements(self):
        """Return the requirements for this param.
        """

        return
        yield

    def get_data(self, context):
        """Get the data.

        Args:
            context (dict): the context (optional) in which the data might be
                found. For Static Param this won't be necessary.
        """

        raise NotImplementedError()


class Param(BaseParam):

    def __init__(self, context_key):
        """Create a default param.

        Args:
            context_key (str): the context key.
        """

        self.context_key = context_key

    @property
    def requirements(self):
        """Return the requirements for this param.
        """

        yield self.context_key

    def get_data(self, context):
        """Get value from the context.

        Args:
            context (dict): the context in which the data might be found based
                on the key provided.

        Return:
            obj: an object from the context that corresponds to the context
                key.
        """

        return context.get(self.context_key, None)


class StaticParam(BaseParam):

    def __init__(self, value):
        """Create a static param.

        Args:
            value (str): the value of the param.
        """

        self.value = value

    def get_data(self, context):
        """Get value from the context.

        Args:
            context (dict): execution context (not used.)
        """

        return self.value


def get_all_requirements(params):
    """Get all the requirements from a list of params.

    Args:
        params (list): The list of params.
    """

    requirements = []
    for param in params:
        requirements += list(param.requirements)
    return requirements


def parametrize(requirement):
    """Parametrize a requirement.

    Args:
        requirement (*): the requirement to parametrize.
    """

    if isinstance(requirement, str):
        return Param(requirement)
    elif isinstance(requirement, BaseParam):
        return requirement
    raise UnknownParamException()


class UnknownParamException(Exception):
    pass



try:
    from unittest.mock import MagicMock
except:
    from mock import MagicMock
from botocore import exceptions
import datetime
import pytest
import json

from garcon import utils


def test_create_dictionary_key_with_string():
    """Try creating a key using a string instead of a dict.
    """

    with pytest.raises(TypeError):
        utils.create_dictionary_key('something')


def test_create_dictionary_key_with_empty_dict():
    """Try creating a key using an empty dictionary.
    """

    with pytest.raises(ValueError):
        utils.create_dictionary_key(dict())


def test_create_dictionary_key():
    """Try creating a unique key from a dict.
    """

    values = [
        dict(foo=10),
        dict(foo2=datetime.datetime.now())]

    for value in values:
        assert len(utils.create_dictionary_key(value)) == 40

def test_create_dictionary_key():
    """Try creating a unique key from a dict.
    """

    values = [
        dict(foo=10),
        dict(foo2=datetime.datetime.now())]

    for value in values:
        assert len(utils.create_dictionary_key(value)) == 40


def test_non_throttle_error():
    """Assert SWF error is evaluated as non-throttle error properly.
    """

    exception = exceptions.ClientError(
        {'Error': {'Code': 'ThrottlingException'}},
        'operationName')
    result = utils.non_throttle_error(exception)
    assert not utils.non_throttle_error(exception)

    exception = exceptions.ClientError(
        {'Error': {'Code': 'RateExceeded'}},
        'operationName')
    assert utils.non_throttle_error(exception)

def test_throttle_backoff_handler():
    """Assert backoff is logged correctly.
    """

    mock_activity = MagicMock()
    details = dict(
        args=(mock_activity,),
        tries=5,
        wait=10)
    utils.throttle_backoff_handler(details)
    mock_activity.logger.info.assert_called_with(
        'Throttle Exception occurred on try {}. '
        'Sleeping for {} seconds'.format(
            details['tries'], details['wait']))



import pytest

from garcon import param


def test_base_param_class():
    """Test the base class: cannot get data and return no requirements.
    """

    current_param = param.BaseParam()
    with pytest.raises(NotImplementedError):
        current_param.get_data({})

    assert not list(current_param.requirements)


def test_static_param():
    """Test the behavior of the static param class
    """

    message = 'Hello World'
    current_param = param.StaticParam(message)
    assert current_param.get_data({}) is message
    assert not list(current_param.requirements)


def test_default_param():
    """Test the behavior of the default param class.
    """

    key = 'context.key'
    message = 'Hello World'
    current_param = param.Param(key)
    requirements = list(current_param.requirements)
    assert current_param.get_data({key: message}) is message
    assert requirements[0] is key


def test_all_requirements():
    """Test getting all the requirements.
    """

    keys = ['context.key1', 'context.key2', 'context.key3']
    manual_keys = ['context.manual_key1', 'context.manual_key2']
    params = [param.Param(key) for key in keys]
    params += manual_keys
    params += [param.StaticParam('Value')]
    params = [param.parametrize(current_param) for current_param in params]

    resp = param.get_all_requirements(params)
    for key in keys:
        assert key in resp

    for manual_key in manual_keys:
        assert manual_key in resp

    assert 'Value' not in resp


def test_parametrize():
    """Test parametrize.

    Parametrize only allows objects that inherits BaseParam or string.
    """

    keys = ['context.key1', 'context.key2', 'context.key3']
    manual_keys = ['context.manual_key1', 'context.manual_key2']
    params = [param.Param(key) for key in keys]
    params += manual_keys
    params += [param.StaticParam('Value')]
    params = [param.parametrize(current_param) for current_param in params]

    for current_param in params:
        assert isinstance(current_param, param.BaseParam)

    with pytest.raises(param.UnknownParamException):
        param.parametrize(list('Unknown'))



from unittest.mock import MagicMock

from garcon import context
from tests.fixtures import decider as decider_events


def test_context_creation_without_events(monkeypatch):
    """Check the basic context creation.
    """

    current_context = context.ExecutionContext()
    assert not current_context.current
    assert not current_context.workflow_input


def test_context_creation_with_events(monkeypatch):
    """Test context creation with events.
    """

    from tests.fixtures import decider as poll

    current_context = context.ExecutionContext(poll.history.get('events'))
    assert current_context.current == {'k': 'v'}

def test_get_workflow_execution_info(monkeypatch):
    """Check that the workflow execution info are properly extracted
    """

    from tests.fixtures import decider as poll

    current_context = context.ExecutionContext()
    current_context.set_workflow_execution_info(poll.history, 'dev')

    # Test extracting workflow execution info
    assert current_context.current == {
        'execution.domain': 'dev',
        'execution.run_id': '123abc=',
        'execution.workflow_id': 'test-workflow-id'}



"""Garcon logger module
"""

import logging


LOGGER_PREFIX = 'garcon'


class GarconLogger:
    """This class is meant to be extended to get the Garcon logger feature
    The logger injects the execution context into the logger name.

    This is used by the Activity class in Garcon and allows you to log from
    the activity object. Typically, you can log from a Garcon task and it will
    prefix your log messages with execution context information (domain,
    workflow_id, run_id).

    Requirements:
    Your loggers need to be set up so there is at least one of them with a name
    prefixed with LOGGER_PREFIX. The Garcon logger will inherit the handlers
    from that logger.

    The formatter for your handler(s) must use the logger name.
    Formatter Example::

        %(asctime)s - %(name)s - %(levelname)s - %(message)s

    This formatter will generate a log message as follow:
    '2015-01-15 - garcon.[domain].[workflow_id].[run_id] - [level] - [message]'
    """

    @property
    def logger(self):
        """Return the appropriate logger. Default to LOGGER_PREFIX if
        no logger name was set.

        Return:
            logging.Logger: a logger object
        """

        return logging.getLogger(
            getattr(self, 'logger_name', None) or LOGGER_PREFIX)

    def set_log_context(self, execution_context):
        """Set a logger name with execution context passed in.

        Args:
            execution_context (dict): execution context information
        """

        if ('execution.domain' in execution_context and
                'execution.workflow_id' in execution_context and
                'execution.run_id' in execution_context):

            self.logger_name = get_logger_namespace(execution_context)

    def unset_log_context(self):
        """Unset the logger name.
        """

        self.logger_name = None


def get_logger_namespace(execution_context):
    """Return the logger namespace for a given execution context.

    Args:
        execution_context (dict): execution context information
    """

    return '.'.join([
        LOGGER_PREFIX,
        execution_context.get('execution.domain'),
        execution_context.get('execution.workflow_id'),
        execution_context.get('execution.run_id')])



import functools
from unittest.mock import MagicMock

import pytest

from garcon import task
from garcon import param


def test_timeout_decorator():
    """Test the timeout decorator.
    """

    timeout = 10

    @task.timeout(timeout)
    def test():
        pass

    assert test.__garcon__.get('timeout') == timeout


def test_timeout_decorator_with_heartbeat():
    """Test the timeout decorator with heartbeat.
    """

    timeout = 20
    heartbeat = 30

    @task.timeout(timeout, heartbeat=heartbeat)
    def test():
        pass

    assert test.__garcon__.get('heartbeat') == heartbeat
    assert test.__garcon__.get('timeout') == timeout

    @task.timeout(timeout)
    @task.decorate(timeout=heartbeat)
    def test2():
        pass

    assert test2.__garcon__.get('heartbeat') == heartbeat
    assert test2.__garcon__.get('timeout') == timeout


def test_decorator():
    """Test the Decorator.

    It should add __garcon__ to the method and if a key / value is
    passed, it should add it.
    """

    def test():
        pass

    task._decorate(test)
    assert hasattr(test, '__garcon__')

    task._decorate(test, 'foo')
    assert test.__garcon__.get('foo') is None

    task._decorate(test, 'foo', 'bar')
    assert test.__garcon__.get('foo') == 'bar'


def test_generator_decorator():
    """Test the geneartor decorator.
    """

    @task.list
    def test():
        pass

    assert test.__garcon__.get('list')
    assert task.is_task_list(test)


def test_link_decorator():
    """Test linking the decorator between two methods.
    """

    def testA():
        pass

    def testB():
        pass

    task._decorate(testA, 'foo', 'value')
    task._link_decorator(testA, testB)
    assert testA.__garcon__ == testB.__garcon__
    assert testA.__garcon__.get('foo') == 'value'
    assert testB.__garcon__.get('foo') == 'value'


def test_link_decorator_with_empty_source():
    """Test linking decorators when garcon is not set on the source.
    """

    def testA():
        pass

    def testB():
        pass

    task._link_decorator(testA, testB)
    assert not getattr(testA, '__garcon__', None)
    assert len(testB.__garcon__) == 0

    task._decorate(testB, 'foo', 'value')
    assert testB.__garcon__.get('foo') == 'value'


def test_task_decorator():
    """Test the task decorator.
    """

    timeout = 40
    userinfo = 'something'

    @task.decorate(timeout=timeout)
    def test(user):
        assert user == userinfo

    assert test.__garcon__.get('timeout') == timeout
    assert test.__garcon__.get('heartbeat') == timeout
    assert callable(test.fill)

    call = test.fill(user='user')
    call(dict(user='something'))


def test_task_decorator_attributes():
    """Test that the decorated task has attributes similar to the original
    function.
    """

    decorator = task.decorate()

    def test():
        pass

    test_task = decorator(test)
    for attr_name in functools.WRAPPER_ASSIGNMENTS:
        assert getattr(test_task, attr_name) == getattr(test, attr_name)


def test_task_decorator_with_heartbeat():
    """Test the task decorator with heartbeat.
    """

    heartbeat = 50
    userinfo = None

    @task.decorate(heartbeat=heartbeat)
    def test(user):
        assert user is userinfo

    assert test.__garcon__.get('heartbeat') == heartbeat


def test_task_decorator_with_activity():
    """Test the task decorator with an activity.
    """

    current_activity = MagicMock()

    @task.decorate()
    def test(activity):
        activity()
        assert activity is current_activity

    call = test.fill()
    call(dict(), activity=current_activity)

    assert current_activity.called


def test_task_decorator_with_context():
    """Test the task decorator with an activity.
    """

    current_context = {}
    spy = MagicMock()

    @task.decorate()
    def test(context):
        spy()

    call = test.fill()

    with pytest.raises(Exception):
        call(current_context)

    assert not spy.called


def test_contextify_added_fill():
    """Verify that calling contextify on a method added .fill on the method.
    """

    @task.contextify
    def test(activity, context, something):
        pass

    assert test
    assert test.fill


def test_contextify_default_method_call():
    """Test that the contextify decorator is not altering the method itself.
    The contextify decorator should preserve the format of the method, it
    allows us to use it in more than one context.
    """

    response = dict(somekey='somevalue')
    spy = MagicMock()

    @task.contextify
    def method(activity, key_to_replace, key_to_not_replace=None):
        assert key_to_replace == 'value'
        assert key_to_not_replace is None
        spy()
        return response

    assert method(None, 'value') is response
    assert spy.called


def test_contextify():
    """Try using contextify on a method.
    """

    value = 'random'
    kwargs_value = 'more-random'
    return_value = 'a value'
    spy = MagicMock()

    @task.contextify
    def method(
        activity, key_to_replace, key_not_to_replace=None,
            kwargs_to_replace=None):

        assert not key_not_to_replace
        assert key_to_replace == value
        assert kwargs_to_replace == kwargs_value
        spy()
        return dict(return_value=return_value)

    fn = method.fill(
        key_to_replace='context.key',
        kwargs_to_replace='context.kwarg_key')

    resp = fn({'context.key': value, 'context.kwarg_key': kwargs_value})

    assert isinstance(resp, dict)
    assert resp.get('return_value') is return_value


def test_contextify_with_mapped_response():
    """Test the contextify method with mapped response.
    """

    return_value = 'a value'

    @task.contextify
    def method(
        activity, key_to_replace, key_not_to_replace=None,
            kwargs_to_replace=None):

        assert activity == 'activity'
        return dict(return_value=return_value)

    fn = method.fill(
        key_to_replace='context.key',
        kwargs_to_replace='context.kwarg_key',
        namespace='somethingrandom')

    resp = fn(
        {'context.key': 'test', 'context.kwarg_key': 'a'},
        activity='activity')

    assert isinstance(resp, dict)
    assert len(resp) == 1
    assert resp.get('somethingrandom.return_value') is return_value


def test_flatten():
    """Test the flatten function.
    """

    spy = MagicMock

    @task.decorate(timeout=10)
    def task_a(name):
        pass

    @task.decorate(timeout=10)
    def task_b(name):
        pass

    @task.list
    def task_generator(context):
        yield task_b
        if context.get('value'):
            yield task_a

    value = list(task.flatten(
        (task_a, task_b, task_generator, task_a),
        dict(value='something')))
    assert value == [task_a, task_b, task_b, task_a, task_a]


def test_fill_function_call():
    """Test filling the function call.
    """

    def test_function(activity, arg_one, key, kwarg_one=None, kwarg_two=None):
        pass

    requirements = dict(
        arg_one=param.Param('context.arg'),
        kwarg_one=param.Param('context.kwarg'))
    activity = None
    context = {
        'context.arg': 'arg.value',
        'context.kwarg': 'kwarg.value'}

    data = task.fill_function_call(
        test_function, requirements, activity, context)

    assert not data.get('activity')
    assert not data.get('context')
    assert not data.get('key')
    assert not data.get('kwarg_two')
    assert data.get('arg_one') == 'arg.value'
    assert data.get('kwarg_one') == 'kwarg.value'
    test_function(**data)


def test_namespace_result():
    """Test namespacing the results.
    """

    value = 'foo'

    resp = task.namespace_result(dict(test=value), '')
    assert resp.get('test') == value

    resp = task.namespace_result(dict(test=value), 'namespace')
    assert resp.get('namespace.test') == value



# -*- coding: utf-8 -*-
"""
Decider Worker
===============

The decider worker is focused on orchestrating which activity needs to be
executed and when based on the flow procided.
"""

import functools
import json
import uuid

from garcon import activity
from garcon import event
from garcon import log

class DeciderWorker(log.GarconLogger):

    def __init__(self, flow, register=True):
        """Initialize the Decider Worker.

        Args:
            flow (module): Flow module.
            register (boolean): If this flow needs to be register on AWS.
        """

        self.client = flow.client
        self.flow = flow
        self.domain = flow.domain
        self.version = getattr(flow, 'version', '1.0')
        self.activities = activity.find_workflow_activities(flow)
        self.task_list = flow.name
        self.on_exception = getattr(flow, 'on_exception', None)

        if register:
            self.register()

    def get_history(self, identity, poll):
        """Get all the history.

        The full history needs to be recovered from SWF to make sure that all
        the activities have been properly scheduled. With boto, only the last
        100 events are provided, this methods retrieves all events.

        Args:
            identity (string): The identity of the decider that pulls the information.
            poll (object): The poll object (see AWS SWF for details.)
        Return:
            list: All the events.
        """

        events = poll['events']
        while 'nextPageToken' in poll:
            poll = self.client.poll_for_decision_task(
                domain=self.domain,
                identity=identity,
                taskList=dict(name=self.task_list),
                nextPageToken=poll['nextPageToken'])

            if 'events' in poll:
                events += poll['events']

        # Remove all the events that are related to decisions and only.
        return [e for e in events if not e['eventType'].startswith('Decision')]

    def get_activity_states(self, history):
        """Get the activity states from the history.

        From the full history extract the different activity states. Those
        states contain

        Args:
            history (list): the full history.
        Return:
            dict: list of all the activities and their state. It only contains
                activities that have been scheduled with AWS.
        """

        return event.activity_states_from_events(history)

    def register(self):
        """Register the Workflow on SWF.

        To work, SWF needs to have pre-registered the domain, the workflow,
        and the different activities, this method takes care of this part.
        """

        registerables = []
        registerables.append((
            self.client.register_domain,
            dict(name=self.domain,
                 workflowExecutionRetentionPeriodInDays='90')))
        registerables.append((
            self.client.register_workflow_type,
            dict(
                domain=self.domain,
                name=self.task_list,
                version=self.version)))

        for current_activity in self.activities:
            registerables.append((
                self.client.register_activity_type,
                dict(
                    domain=self.domain,
                    name=current_activity.name,
                    version=self.version)))

        for callable, data in registerables:
            try:
                callable(**data)
            except Exception as e:
                print(e)
                print(data.get('name'), 'already exists')

    def create_decisions_from_flow(self, decisions, activity_states, context):
        """Create the decisions from the flow.

        Simple flows don't need a custom decider, since all the requirements
        can be provided at the activity level. Discovery of the next activity
        to schedule is thus very straightforward.

        Args:
            decisions (list): the layer decision for swf.
            activity_states (dict): all the state activities.
            context (dict): the context of the activities.
        """

        try:
            for current in activity.find_available_activities(
                    self.flow, activity_states, context.current):
                schedule_activity_task(
                    decisions, current, version=self.version)
            else:
                activities = list(
                    activity.find_uncomplete_activities(
                        self.flow, activity_states, context.current))
                if not activities:
                    decisions.append(dict(
                        decisionType='CompleteWorkflowExecution'))
        except Exception as e:
            decisions.append(dict(
                decisionType='FailWorkflowExecution',
                failWorkflowExecutionDecisionAttributes=dict(
                    reason=str(e))))
            if self.on_exception:
                self.on_exception(self, e)
            self.logger.error(e, exc_info=True)

    def delegate_decisions(self, decisions, decider, history, context):
        """Delegate the decisions.

        For more complex flows (the ones that have, for instance, optional
        activities), you can write your own decider. The decider receives a
        method `schedule` which schedule the activity if not scheduled yet,
        and if scheduled, returns its result.

        Args:
            decisions (list): the layer decision for swf.
            decider (callable): the decider (it needs to have schedule)
            history (dict): all the state activities and its history.
            context (dict): the context of the activities.
        """

        schedule_context = ScheduleContext()
        decider_schedule = functools.partial(
            schedule, decisions, schedule_context, history, context.current,
            version=self.version)

        try:
            kwargs = dict(schedule=decider_schedule)

            # retro-compatibility.
            if 'context' in decider.__code__.co_varnames:
                kwargs.update(context=context.workflow_input)

            decider(**kwargs)

            # When no exceptions are raised and the method decider has returned
            # it means that there i nothing left to do in the current decider.
            if schedule_context.completed:
                decisions.append(dict(
                    decisionType='CompleteWorkflowExecution'))
        except activity.ActivityInstanceNotReadyException:
            pass
        except Exception as e:
            decisions.append(dict(
                decisionType='FailWorkflowExecution',
                failWorkflowExecutionDecisionAttributes=dict(
                    reason=str(e))))
            if self.on_exception:
                self.on_exception(self, e)
            self.logger.error(e, exc_info=True)

    def run(self, identity=None):
        """Run the decider.

        The decider defines which task needs to be launched and when based on
        the list of events provided. It looks at the list of all the available
        activities, and launch the ones that:

          * are not been scheduled yet.
          * have all the dependencies resolved.

        If the decider is not able to find an uncompleted activity, the
        workflow can safely mark its execution as complete.

        Args:
            identity (str): Identity of the worker making the request, which
                is recorded in the DecisionTaskStarted event in the AWS
                console. This enables diagnostic tracing when problems arise.

        Return:
            boolean: Always return true, so any loop on run can act as a long
                running process.
        """

        try:
            poll = self.client.poll_for_decision_task(
                domain=self.domain,
                taskList=dict(name=self.task_list),
                identity=identity or '')
        except Exception as error:
            # Catch exceptions raised during poll() to avoid a Decider thread
            # dying & the daemon unable to process subsequent workflows.
            # AWS api limits on SWF calls are a common source of such
            # exceptions.

            # on_exception() can be overriden by the flow to send an alert
            # when such an exception occurs.
            if self.on_exception:
                self.on_exception(self, error)
            self.logger.error(error, exc_info=True)
            return True

        custom_decider = getattr(self.flow, 'decider', None)

        if 'events' not in poll:
            return True

        history = self.get_history(identity or '', poll)
        activity_states = self.get_activity_states(history)
        current_context = event.get_current_context(history)
        current_context.set_workflow_execution_info(poll, self.domain)

        decisions = []
        if not custom_decider:
            self.create_decisions_from_flow(
                decisions, activity_states, current_context)
        else:
            self.delegate_decisions(
                decisions, custom_decider, activity_states, current_context)
        self.client.respond_decision_task_completed(
            taskToken=poll.get('taskToken'),
            decisions=decisions)
        return True


class ScheduleContext:
    """
    Schedule Context
    ================

    The schedule context keeps track of all the current scheduling progress –
    which allows to easy determinate if there are more decisions to be taken
    or if the execution can be closed.
    """

    def __init__(self):
        """Create a schedule context.
        """

        self.completed = True

    def mark_uncompleted(self):
        """Mark the scheduling as completed.

        When a scheduling is completed, it means all the activities have been
        properly scheduled and they have all completed.
        """

        self.completed = False


def schedule_activity_task(
        decisions, instance, version='1.0', id=None):
    """Schedule an activity task.

    Args:
        decisions (Layer1Decisions): the layer decision for swf.
        instance (ActivityInstance): the activity instance to schedule.
        version (str): the version of the activity instance.
        id (str): optional id of the activity instance.
    """

    decisions.append(dict(
        decisionType='ScheduleActivityTask',
        scheduleActivityTaskDecisionAttributes=dict(
            activityId=id or instance.id,
            activityType=dict(
                name=instance.activity_name,
                version=version),
            taskList=dict(name=instance.activity_worker.task_list),
            input=json.dumps(instance.create_execution_input()),
            heartbeatTimeout=str(instance.heartbeat_timeout),
            startToCloseTimeout=str(instance.timeout),
            scheduleToStartTimeout=str(instance.schedule_to_start),
            scheduleToCloseTimeout=str(instance.schedule_to_close))))


def schedule(
        decisions, schedule_context, history, context, schedule_id,
        current_activity, requires=None, input=None, version='1.0'):
    """Schedule an activity.

    Scheduling an activity requires all the requirements to be completed (all
    activities should be marked as completed). The scheduler also mixes the
    input with the full execution context to send the data to the activity.

    Args:
        decisions (list): the layer decision for swf.
        schedule_context (dict): information about the schedule.
        history (dict): history of the execution.
        context (dict): context of the execution.
        schedule_id (str): the id of the activity to schedule.
        current_activity (Activity): the activity to run.
        requires (list): list of all requirements.
        input (dict): additional input for the context.

    Throws:
        ActivityInstanceNotReadyException: if one of the activity in the
            requirements is not ready.

    Return:
        State: the state of the schedule (contains the response).
    """

    ensure_requirements(requires)
    activity_completed = set()
    result = dict()

    instance_context = dict()
    instance_context.update(context or {})
    instance_context.update(input or {})

    for current in current_activity.instances(instance_context):
        current_id = '{}-{}'.format(current.id, schedule_id)
        states = history.get(current.activity_name, {}).get(current_id)

        if states:
            if states.get_last_state() == activity.ACTIVITY_COMPLETED:
                result.update(states.result or dict())
                activity_completed.add(True)
                continue

            activity_completed.add(False)
            schedule_context.mark_uncompleted()

            if states.get_last_state() != activity.ACTIVITY_FAILED:
                continue
            elif (not current.retry or
                  current.retry < activity.count_activity_failures(states)):
                raise Exception(
                    'The activity failures has exceeded its retry limit.')

        activity_completed.add(False)
        schedule_context.mark_uncompleted()
        schedule_activity_task(
            decisions, current, id=current_id, version=version)

    state = activity.ActivityState(current_activity.name)
    state.add_state(activity.ACTIVITY_SCHEDULED)

    if len(activity_completed) == 1 and True in activity_completed:
        state.add_state(activity.ACTIVITY_COMPLETED)
        state.set_result(result)
    return state


def ensure_requirements(requires):
    """Ensure scheduling meets requirements.

    Verify the state of the requirements to make sure the activity can be
    scheduled.

    Args:
        requires (list): list of all requirements.

    Throws:
        ActivityInstanceNotReadyException: if one of the activity in the
            requirements is not ready.
    """

    requires = requires or []
    for require in requires:
        if (not require or
                require.get_last_state() != activity.ACTIVITY_COMPLETED):
            raise activity.ActivityInstanceNotReadyException()



# -*- coding: utf-8 -*-
"""
Context
=======

Context carries information that have been retrieved from the different SWF
events of an execution.
"""

import json


class ExecutionContext:

    def __init__(self, events=None):
        """Create the execution context.

        An execution context gathers the execution input and the result of all
        the activities that have successfully ran. It also adds the execution
        input into the mix (for logger purposes).

        Args:
            events (list): optional list of all the events.
        """

        self.current = {}
        self.workflow_input = {}

        if events:
            for event in events:
                self.add(event)

    def add(self, event):
        """Add an event into the execution context.

        The events are the ones coming from SWF directly (so the fields are the
        ones we expect).

        Args:
            event (dict): the event to add to the context.
        """

        event_type = event.get('eventType')
        if event_type == 'ActivityTaskCompleted':
            self.add_activity_result(event)
        elif event_type == 'WorkflowExecutionStarted':
            self.set_execution_input(event)

    def set_workflow_execution_info(self, execution_info, domain):
        """Add the workflow execution info.

        Workflow execution info contains the domain, workflow id and run id.
        This allows the logger to properly namespace the messages and
        facilitate debugging.

        Args:
            execution_info (dict): the execution information.
            domain (str): the current domain
        """

        if ('workflowExecution' in execution_info and
                'workflowId' in execution_info['workflowExecution'] and
                'runId' in execution_info['workflowExecution']):

            workflow_execution = execution_info['workflowExecution']
            self.current.update({
                'execution.domain': domain,
                'execution.workflow_id': workflow_execution['workflowId'],
                'execution.run_id': workflow_execution['runId']
            })

    def set_execution_input(self, execution_event):
        """Add the workflow execution input.

        Please note the input within the execution event should always be a
        json string.

        Args:
            execution_event (str): the execution event information.
        """

        attributes = execution_event['workflowExecutionStartedEventAttributes']
        result = attributes.get('input')
        if result:
            result = json.loads(result)
            self.workflow_input = result
            self.current.update(result)

    def add_activity_result(self, activity_event):
        """Add an activity result.

        Please note: the result of an activity event should always be a json
        string.

        Args:
            activity_event (str): json object that represents the activity
                information.
        """

        attributes = activity_event['activityTaskCompletedEventAttributes']
        result = attributes.get('result')

        if result:
            self.current.update(json.loads(result))



# -*- coding: utf-8 -*-
"""
Activity
========

Activities are self generated classes to which you can pass an identifier,
and a list of tasks to perform. The activities are in between the decider and
the tasks.

For ease, two types of task runners are available: Sync and Async. If
you need something more specific, you should either create your own runner, or
you should create a main task that will then split the work.

Create an activity::

    import boto3
    from garcon import activity

    # First step is to create the workflow on a specific domain.
    client = boto3.client('swf')
    create = activity.create(client, 'domain', 'workflow-name')

    initial_activity = create(
        # Name of your activity
        name='activity_name',

        # List of tasks to run (here we use the Sync runner)
        run=runner.Sync(task1),

        # No requires since it's the first one. Later in your flow, if you have
        # a dependency, just use the variable that contains the activity.
        requires=[],

        # If the activity fails, number of times you want to retry.
        retry=0,

        # If you want to run the activity `n` times, you can use a generator.
        generator=[generator_name])

"""

from botocore import exceptions
import itertools
import json
import threading
import backoff

from garcon import log
from garcon import utils
from garcon import runner

ACTIVITY_STANDBY = 0
ACTIVITY_SCHEDULED = 1
ACTIVITY_COMPLETED = 2
ACTIVITY_FAILED = 3

DEFAULT_ACTIVITY_SCHEDULE_TO_START = 600  # 10 minutes


class ActivityInstanceNotReadyException(Exception):
    """Exception when an activity instance is not ready.

    Activity instances that are considered not ready are instances that have
    not completed.
    """

    pass


class ActivityInstance:

    def __init__(
            self, activity_worker, local_context=None, execution_context=None):
        """Activity Instance.

        In SWF, Activity is a worker: it will get information from the context,
        and will launch activity instances (only one, unless you have a
        generator.) The activity instance generates its key (visible in the SWF
        console) from the local context. Activity instances are owned by an
        execution.

        Args:
            activity_worker (ActivityWorker): The activity worker that owns
                this specific Activity Instance.
            local_context (dict): the local context of the activity (it does
                not include the execution context.) Most times the context will
                be empty since it is only filled with data that comes from the
                generators.
            execution_context (dict): the execution context of when an activity
                will be scheduled with.
        """

        self.activity_worker = activity_worker
        self.execution_context = execution_context or dict()
        self.local_context = local_context or dict()
        self.global_context = dict(
            list(self.execution_context.items()) +
            list(self.local_context.items()))

    @property
    def activity_name(self):
        """Return the activity name of the worker.
        """

        return self.activity_worker.name

    @property
    def retry(self):
        """Return the number of retries allowed (matches the worker.)
        """

        return self.activity_worker.retry

    @property
    def id(self):
        """Generate the id of the activity.

        The id is crutial (not just important): it allows to indentify the
        state the activity instance in the event history (if it has failed,
        been executed, or marked as completed.)

        Return:
            str: composed of the activity name (task list), and the activity
                id.
        """

        if not self.local_context:
            activity_id = 1
        else:
            activity_id = utils.create_dictionary_key(self.local_context)

        return '{name}-{id}'.format(
            name=self.activity_name,
            id=activity_id)

    @property
    def schedule_to_start(self):
        """Return the schedule to start timeout.

        The schedule to start timeout assumes that only one activity worker is
        available (since swf does not provide a count of available workers). So
        if the default value is 5 minutes, and you have 10 instances: the
        schedule to start will be 50 minutes for all instances.

        Return:
            int: Schedule to start timeout.
        """

        return (
            self.activity_worker.pool_size *
            self.activity_worker.schedule_to_start_timeout)

    @property
    def schedule_to_close(self):
        """Return the schedule to close timeout.

        The schedule to close timeout is a simple calculation that defines when
        an activity (from the moment it has been scheduled) should end. It is
        a calculation between the schedule to start timeout and the activity
        timeout.

        Return:
            int: Schedule to close timeout.
        """

        return self.schedule_to_start + self.timeout

    @property
    def timeout(self):
        """Return the timeout in seconds.

        This timeout corresponds on when the activity has started and when we
        assume the activity has ended (which corresponds in boto to
        start_to_close_timeout.)

        Return:
            int: Task list timeout.
        """

        return self.runner.timeout(self.global_context)

    @property
    def heartbeat_timeout(self):
        """Return the heartbeat in seconds.

        This heartbeat corresponds on when an activity needs to send a signal
        to swf that it is still running. This will set the value when the
        activity is scheduled.

        Return:
            int: Task list timeout.
        """

        return self.runner.heartbeat(self.global_context)

    @property
    def runner(self):
        """Shortcut to get access to the runner.

        Raises:
            runner.RunnerMissing: an activity should always have a runner,
                if the runner is missing an exception is raised (we will not
                be able to calculate values such as timeouts without a runner.)

        Return:
            Runner: the activity runner.
        """

        activity_runner = getattr(self.activity_worker, 'runner', None)
        if not activity_runner:
            raise runner.RunnerMissing()
        return activity_runner

    def create_execution_input(self):
        """Create the input of the activity from the context.

        AWS has a limit on the number of characters that can be used (32k). If
        you use the `task.decorate`, the data sent to the activity is optimized
        to match the values of the context as well as the execution context.

        Return:
            dict: the input to send to the activity.
        """

        activity_input = dict()

        try:
            for requirement in self.runner.requirements(self.global_context):
                value = self.global_context.get(requirement)
                if value is not None:
                    activity_input.update({requirement: value})

            activity_input.update({
                'execution.domain': self.global_context.get('execution.domain'),
                'execution.run_id': self.global_context.get('execution.run_id'),
                'execution.workflow_id': self.global_context.get(
                    'execution.workflow_id')
            })

        except runner.NoRunnerRequirementsFound:
            return self.global_context

        return activity_input


class Activity(log.GarconLogger):
    version = '1.0'
    task_list = None

    def __init__(self, client):
        """Instantiates an activity.

        Args:
            client: the boto client used for this activity.
        """

        self.client = client
        self.name = None
        self.domain = None
        self.task_list = None

    @backoff.on_exception(
        backoff.expo,
        exceptions.ClientError,
        max_tries=5,
        giveup=utils.non_throttle_error,
        on_backoff=utils.throttle_backoff_handler,
        jitter=backoff.full_jitter)
    def poll_for_activity(self, identity=None):
        """Runs Activity Poll.

        If a SWF throttling exception is raised during a poll, the poll will
        be retried up to 5 times using exponential backoff algorithm.

        Upgrading to boto3 would make this retry logic redundant.

        Args:
            identity (str): Identity of the worker making the request, which
                is recorded in the ActivityTaskStarted event in the AWS
                console. This enables diagnostic tracing when problems arise.
        Return:
            ActivityExecution: activity execution.
        """

        additional_params = {}
        if identity:
            additional_params.update(identity=identity)

        execution_definition = self.client.poll_for_activity_task(
            domain=self.domain, taskList=dict(name=self.task_list),
            **additional_params)

        return ActivityExecution(
            self.client, execution_definition.get('activityId'),
            execution_definition.get('taskToken'),
            execution_definition.get('input'))

    def run(self, identity=None):
        """Activity Runner.

        Information is being pulled down from SWF and it checks if the Activity
        can be ran. As part of the information provided, the input of the
        previous activity is consumed (context).

        Args:
            identity (str): Identity of the worker making the request, which
                is recorded in the ActivityTaskStarted event in the AWS
                console. This enables diagnostic tracing when problems arise.
        """
        try:
            if identity:
                self.logger.debug('Polling with {}'.format(identity))
            execution = self.poll_for_activity(identity)
        except Exception as error:
            # Catch exceptions raised during poll() to avoid an Activity thread
            # dying & worker daemon unable to process the affected Activity.
            # AWS api limits on SWF calls are a common source of such
            # exceptions (see https://github.com/xethorn/garcon/pull/75)

            # on_exception() can be overriden by the flow to send an alert
            # when an exception occurs.
            if self.on_exception:
                self.on_exception(self, error)
            self.logger.error(error, exc_info=True)
            return True

        self.set_log_context(execution.context)
        if execution.activity_id:
            try:
                context = self.execute_activity(execution)
                execution.complete(context)
            except Exception as error:
                # If the workflow has been stopped, it is not possible for the
                # activity to be updated – it throws an exception which stops
                # the worker immediately.
                try:
                    execution.fail(str(error)[:255])
                    if self.on_exception:
                        self.on_exception(self, error)
                except Exception as error2:  # noqa: E722
                    if self.on_exception:
                        self.on_exception(self, error2)

        self.unset_log_context()
        return True

    def execute_activity(self, activity):
        """Execute the runner.

        Args:
            execution (ActivityExecution): the activity execution.

        Return:
            dict: The result of the operation.
        """

        return self.runner.execute(activity, activity.context)

    def hydrate(self, data):
        """Hydrate the task with information provided.

        Args:
            data (dict): the data to use (if defined.)
        """

        self.pool_size = 0
        self.version = data.get('version') or self.version
        self.name = self.name or data.get('name')
        self.domain = getattr(self, 'domain', '') or data.get('domain')
        self.requires = getattr(self, 'requires', []) or data.get('requires')
        self.retry = getattr(self, 'retry', None) or data.get('retry', 0)
        self.task_list = self.task_list or data.get('task_list')
        self.on_exception = (
            getattr(self, 'on_exception', None) or data.get('on_exception'))

        # The start timeout is how long it will take between the scheduling
        # of the activity and the start of the activity.
        self.schedule_to_start_timeout = (
            getattr(self, 'schedule_to_start_timeout', None) or
            data.get('schedule_to_start') or
            DEFAULT_ACTIVITY_SCHEDULE_TO_START)

        # The previous way to create an activity was to fill a `tasks` param,
        # which is not `run`.
        self.runner = (
            getattr(self, 'runner', None) or
            data.get('run') or data.get('tasks'))

        self.generators = getattr(
            self, 'generators', None) or data.get('generators')

    def instances(self, context):
        """Get all instances for one activity based on the current context.

        There are two scenarios: when the activity worker has a generator and
        when it does not. When it doesn't (the most simple case), there will
        always be one instance returned.

        Generators will however consume the context to calculate how many
        instances of the activity are needed – and it will generate them
        (regardless of their state.)

        Args:
            context (dict): the current context.
        Return:
            list: all the instances of the activity (for a current workflow
                execution.)
        """

        if not self.generators:
            self.pool_size = 1
            yield ActivityInstance(self, execution_context=context)
            return

        generator_values = []
        for generator in self.generators:
            generator_values.append(generator(context))

        contexts = list(itertools.product(*generator_values))
        self.pool_size = len(contexts)
        for generator_contexts in contexts:
            # Each generator returns a context, merge all the contexts
            # to only be one - which can be used to 1/ create the id of the
            # activity and 2/ be passed as a local context.
            instance_context = dict()
            for current_generator_context in generator_contexts:
                instance_context.update(current_generator_context.items())

            yield ActivityInstance(
                self, execution_context=context,
                local_context=instance_context)


class ExternalActivity(Activity):
    """External activity

    One of the main advantages of SWF is the ability to write a workflow that
    has activities written in any languages. The external activity class allows
    to write the workflow in Garcon and benefit from some features (timeout
    calculation among other things, sending context data.)
    """

    def __init__(self, timeout=None, heartbeat=None):
        """Create the External Activity.

        Args:
            timeout (int): activity timeout in seconds (mandatory)
            heartbeat (int): heartbeat timeout in seconds, if not defined, it
                will be equal to the timeout.
        """

        Activity.__init__(self, client=None)
        self.runner = runner.External(timeout=timeout, heartbeat=heartbeat)

    def run(self):
        """Run the external activity.

        This activity is handled outside, so the run method should remain
        unimplemented and return False (so the run loop stops.)
        """

        return False


class ActivityExecution(log.GarconLogger):

    def __init__(self, client, activity_id, task_token, context):
        """Create an an activity execution.

        Args:
            client (boto3.client): the boto client (for easy access if needed).
            activity_id (str): the activity id.
            task_token (str): the task token.
            context (str): data for the execution.
        """

        self.client = client
        self.activity_id = activity_id
        self.task_token = task_token
        self.context = context and json.loads(context) or dict()

    def heartbeat(self, details=None):
        """Create a task heartbeat.

        Args:
            details (str): details to add to the heartbeat.
        """

        self.client.record_activity_task_heartbeat(taskToken=self.task_token,
            details=details or '')

    def fail(self, reason=None):
        """Mark the activity execution as failed.

        Args:
            reason (str): optional reason for the failure.
        """

        self.client.respond_activity_task_failed(
            taskToken=self.task_token,
            reason=reason or '')

    def complete(self, context=None):
        """Mark the activity execution as completed.

        Args:
            context (str or dict): the context result of the operation.
        """

        self.client.respond_activity_task_completed(
            taskToken=self.task_token,
            result=json.dumps(context))


class ActivityWorker():

    def __init__(self, flow, activities=None):
        """Initiate an activity worker.

        The activity worker take in consideration all the activities from a
        flow, or specific activities. Some activities (tasks) might require
        more power than others, and be then launched on different machines.

        If a list of activities is passed, the worker will be focused on
        completing those and will ignore all the others.

        Args:
            flow (module): the flow module.
            activities (list): the list of activities that this worker should
                handle.
        """

        self.flow = flow
        self.activities = find_workflow_activities(self.flow)
        self.worker_activities = activities

    def run(self):
        """Run the activities.
        """

        for activity in self.activities:
            if (self.worker_activities and
                    activity.name not in self.worker_activities):
                continue
            threading.Thread(target=worker_runner, args=(activity,)).start()


class ActivityState:
    """
    Activity State
    ==============

    Provides information about a specific activity instance state (if the
    instance is already scheduled, has failed, or has been completed.) Along
    with the default values, this class also provides additional metadata such
    as the result of an activity instance.
    """

    def __init__(self, activity_id):
        """Create a State.

        Args:
            activity_id (str): the activity id.
        """

        self.activity_id = activity_id
        self._result = None
        self.states = []

    @property
    def result(self):
        """Get the result.
        """

        if not self.ready:
            raise ActivityInstanceNotReadyException()
        return self._result

    @property
    def ready(self):
        """Check if an activity is ready.
        """

        return self.get_last_state() == ACTIVITY_COMPLETED

    def get_last_state(self):
        """Get the last state of the activity execution.

        Return:
            int: the state of the activity (see: activity.py)
        """

        if len(self.states):
            return self.states[-1]
        return None

    def add_state(self, state):
        """Add a state in the activity execution.

        Args:
            state (int): the state of the activity to add (see activity.py)
        """

        self.states.append(state)

    def set_result(self, result):
        """Set the result of the activity.

        This method sometimes throws an exception: an activity id can only have
        one result.

        Args:
            result (dict): Result of the activity.
        """

        if self._result:
            raise Exception('Result is ummutable – it should not be changed.')
        self._result = result

    def wait(self):
        """Wait until ready.
        """

        if not self.ready:
            raise ActivityInstanceNotReadyException()


def worker_runner(worker):
    """Run indefinitely the worker.

    Args:
        worker (object): the Activity worker.
    """

    while (worker.run()):
        continue


def create(client, domain, workflow_name, version='1.0', on_exception=None):
    """Helper method to create Activities.

    The helper method simplifies the creation of an activity by setting the
    domain, the task list, and the activity dependencies (what other
    activities) need to be completed before this one can run.

    Note:
        The task list is generated based on the domain and the name of the
        activity. Always make sure your activity name is unique.

    Args:
        client (boto3.client): the boto3 client.
        domain (str): the domain name.
        workflow_name (str): workflow name.
        version (str): activity version.
        on_exception (callable): the error handler.

    Return:
        callable: activity generator.
    """

    def wrapper(**options):
        activity = Activity(client)

        if options.get('external'):
            activity = ExternalActivity(
                timeout=options.get('timeout'),
                heartbeat=options.get('heartbeat'))

        activity_name = '{name}_{activity}'.format(
            name=workflow_name,
            activity=options.get('name'))

        activity.hydrate(dict(
            domain=domain,
            version=version,
            name=activity_name,
            generators=options.get('generators', []),
            requires=options.get('requires', []),
            retry=options.get('retry'),
            task_list=activity_name,
            tasks=options.get('tasks'),
            run=options.get('run'),
            schedule_to_start=options.get('schedule_to_start'),
            on_exception=options.get('on_exception') or on_exception))
        return activity

    return wrapper


def find_available_activities(flow, history, context):
    """Find all available activity instances of a flow.

    The history contains all the information of our activities (their state).
    This method focuses on finding all the activities that need to run.

    Args:
        flow (module): the flow module.
        history (dict): the history information.
        context (dict): from the context find the available activities.
    """

    for instance in find_activities(flow, context):
        # If an event is already available for the activity, it means it is
        # not in standby anymore, it's either processing or has been completed.
        # The activity is thus not available anymore.
        states = history.get(instance.activity_name, {}).get(instance.id)

        if states:
            if states.get_last_state() != ACTIVITY_FAILED:
                continue
            elif (not instance.retry or
                  instance.retry < count_activity_failures(states)):
                raise Exception(
                    'The activity failures has exceeded its retry limit.')

        can_yield = True
        for requirement in instance.activity_worker.requires:
            require_history = history.get(requirement.name)

            if not require_history:
                can_yield = False
                break

            for requirement_states in require_history.values():
                if ACTIVITY_COMPLETED not in requirement_states.states:
                    can_yield = False
                    break

        if can_yield:
            yield instance


def find_uncomplete_activities(flow, history, context):
    """Find uncomplete activity instances.

    Uncomplete activities are all the activities that are not marked as
    completed.

    Args:
        flow (module): the flow module.
        history (dict): the history information.
        context (dict): from the context find the available activities.
    Yield:
        activity: The available activity.
    """

    for instance in find_activities(flow, context):
        states = history.get(instance.activity_name, {}).get(instance.id)
        if not states or ACTIVITY_COMPLETED not in states.states:
            yield instance


def find_workflow_activities(flow):
    """Retrieves all the activities from a flow

    Args:
        flow (module): the flow module.
    Return:
        list: all the activities.
    """

    activities = []
    for module_attribute in dir(flow):
        current_activity = getattr(flow, module_attribute)
        if isinstance(current_activity, Activity):
            activities.append(current_activity)
    return activities


def find_activities(flow, context):
    """Retrieves all the activities from a flow.

    Args:
        flow (module): the flow module.
    Return:
        list: All the activity instances for the flow.
    """

    activities = []
    for module_attribute in dir(flow):
        current_activity = getattr(flow, module_attribute)

        if isinstance(current_activity, Activity):
            for activity_instance in current_activity.instances(context):
                activities.append(activity_instance)

    return activities


def count_activity_failures(states):
    """Count the number of times an activity has failed.

    Args:
        states (dict): list of activity states.
    Return:
        int: The number of times an activity has failed.
    """

    return len([evt for evt in states.states if evt == ACTIVITY_FAILED])



"""
Utils
=====

"""

import hashlib


def create_dictionary_key(dictionary):
    """Create a key that represents the content of the dictionary.

    Args:
        dictionary (dict): the dictionary to use.
    Return:
        str: the key that represents the content of the dictionary.
    """

    if not isinstance(dictionary, dict):
        raise TypeError('The value passed should be a dictionary.')

    if not dictionary:
        raise ValueError('The dictionary cannot be empty.')

    sorted_dict = sorted(dictionary.items())

    key_parts = ''.join([
        "'{key}':'{val}';".format(key=key, val=val)
        for (key, val) in sorted_dict])

    return hashlib.sha1(key_parts.encode('utf-8')).hexdigest()

def non_throttle_error(exception):
    """Activity Runner.

    Determine whether SWF Exception was a throttle or a different error.

    Args:
        exception: botocore.exceptions.Client instance.
    Return:
        bool: True if exception was a throttle, False otherwise.
    """

    return exception.response.get('Error').get('Code') != 'ThrottlingException'

def throttle_backoff_handler(details):
    """Callback to be used when a throttle backoff is invoked.

    For more details see: https://github.com/litl/backoff/#event-handlers

    Args:
        dictionary (dict): Details of the backoff invocation. Valid keys
            include:
                target: reference to the function or method being invoked.
                args: positional arguments to func.
                kwargs: keyword arguments to func.
                tries: number of invocation tries so far.
                wait: seconds to wait (on_backoff handler only).
                value: value triggering backoff (on_predicate decorator only).
    """

    activity = details['args'][0]
    activity.logger.info(
        'Throttle Exception occurred on try {}. '
        'Sleeping for {} seconds'.format(
            details['tries'], details['wait']))



"""
Task runners
============

The task runners are responsible for running all the tasks (either in series
or in parallel). There's only one task runner per activity.
"""

from concurrent import futures
from concurrent.futures import ThreadPoolExecutor

from garcon.task import flatten


DEFAULT_TASK_TIMEOUT = 600   # 10 minutes.
DEFAULT_TASK_HEARTBEAT = 600  # 10 minutes


class NoRunnerRequirementsFound(Exception):
    pass


class RunnerMissing(Exception):
    pass


class BaseRunner():

    def __init__(self, *args):
        self.tasks = args

    def timeout(self, context):
        """Calculate and return the timeout for an activity.

        The calculation of the timeout is pessimistic: it takes the worse case
        scenario (even for asynchronous task lists, it supposes there is only
        one thread completed at a time.)

        Return:
            str: The timeout (boto requires the timeout to be a string and not
                a regular number.)
        """

        timeout = 0

        for task in flatten(self.tasks, context):
            task_timeout = DEFAULT_TASK_TIMEOUT
            task_details = getattr(task, '__garcon__', None)

            if task_details:
                task_timeout = task_details.get(
                    'timeout', DEFAULT_TASK_TIMEOUT)

            timeout = timeout + task_timeout

        return timeout

    def heartbeat(self, context):
        """Calculate and return the heartbeat for an activity.

        The heartbeat represents when an actvitity should be sending a signal
        to SWF that it has not completed yet. The heartbeat is sent everytime
        a new task is going to be launched.

        Similar to the `BaseRunner.timeout`, the heartbeat is pessimistic, it
        looks at the largest heartbeat and set it up.

        Return:
            str: The heartbeat timeout (boto requires the timeout to be a
                string not a regular number.)
        """

        heartbeat = 0

        for task in flatten(self.tasks, context):
            task_details = getattr(task, '__garcon__', None)
            task_heartbeat = DEFAULT_TASK_HEARTBEAT

            if task_details:
                task_heartbeat = task_details.get(
                    'heartbeat', DEFAULT_TASK_HEARTBEAT)

            if task_heartbeat > heartbeat:
                heartbeat = task_heartbeat

        return heartbeat

    def requirements(self, context):
        """Find all the requirements from the list of tasks and return it.

        If a task does not use the `task.decorate`, no assumptions can be made
        on which values from the context will be used, and it will raise an
        exception.

        Raise:
            NoRequirementFound: The exception when no requirements have been
                mentioned in at least one or more tasks.

        Return:
            set: the list of the required values from the context.
        """

        requirements = []

        # Get all the tasks and the lists (so the .fill on lists are also
        # considered.)
        all_tasks = list(self.tasks) + list(flatten(self.tasks, context))
        for task in all_tasks:
            task_details = getattr(task, '__garcon__', None)
            if task_details:
                requirements += task_details.get('requirements', [])
            else:
                raise NoRunnerRequirementsFound()
        return set(requirements)

    def execute(self, activity, context):
        """Execution of the tasks.
        """

        raise NotImplementedError


class Sync(BaseRunner):

    def execute(self, activity, context):
        result = dict()
        for task in flatten(self.tasks, context):
            activity.heartbeat()
            task_context = dict(list(result.items()) + list(context.items()))
            resp = task(task_context, activity=activity)
            result.update(resp or dict())
        return result


class Async(BaseRunner):

    def __init__(self, *args, **kwargs):
        self.tasks = args
        self.max_workers = kwargs.get('max_workers', 3)

    def execute(self, activity, context):
        result = dict()
        with ThreadPoolExecutor(max_workers=self.max_workers) as executor:
            tasks = []
            for task in flatten(self.tasks, context):
                tasks.append(executor.submit(task, context, activity=activity))

            for future in futures.as_completed(tasks):
                activity.heartbeat()
                data = future.result()
                result.update(data or {})
        return result


class External(BaseRunner):

    def __init__(self, timeout=None, heartbeat=None):
        """Create the External Runner.

        Args:
            timeout (int): activity timeout in seconds (mandatory)
            heartbeat (int): heartbeat timeout in seconds, if not defined, it
                will be equal to the timeout.
        """

        assert timeout, 'External runner requires a timeout.'

        self.timeout = lambda ctx=None: timeout
        self.heartbeat = lambda ctx=None: (heartbeat or timeout)



# -*- coding: utf-8 -*-
"""
Task
====

Tasks are small discrete applications that are meant to perform a defined
action within an activity. An activity can have more than one task, they can
run in series or in parallel.

Tasks can add values to the context by returning a dictionary that contains
the informations to add (useful if you need to pass information from one
task – in an activity, to another activity's task.)

Note:
    If you need a task runner that is not covered by the two scenarios below,
    you may need to just have a main task, and have this task split the work
    the way you want.
"""

import copy
from functools import update_wrapper

from garcon import param


def decorate(timeout=None, heartbeat=None, enable_contextify=True):
    """Generic task decorator for tasks.

    Args:
        timeout (int): The timeout of the task (see timeout).
        heartbeat (int): The heartbeat timeout.
        contextify (boolean): If the task can be contextified (see contextify).
    Return:
        callable: The wrapper.
    """

    def wrapper(fn):
        if timeout:
            _decorate(fn, 'timeout', timeout)

        # If the task does not have a heartbeat, but instead the task has
        # a timeout, the heartbeat should be adjusted to the timeout. In
        # most case, most people will probably opt for this option.
        if heartbeat or timeout:
            _decorate(fn, 'heartbeat', heartbeat or timeout)

        if enable_contextify:
            contextify(fn)
        return fn

    return wrapper


def timeout(time, heartbeat=None):
    """Wrapper for a task to define its timeout.

    Args:
        time (int): the timeout in seconds
        heartbeat (int): the heartbeat timeout (in seconds too.)
    """

    def wrapper(fn):
        _decorate(fn, 'timeout', time)
        if heartbeat:
            _decorate(fn, 'heartbeat', heartbeat)
        return fn

    return wrapper


def list(fn):
    """Wrapper for a callable to define a task generator.

    Generators are used to check values in the context and schedule different
    tasks based on it. Note: depending on the tasks returned by the generator,
    the timout values will be calculated differently.

    For instance::

        @task.list
        def create_client(context):
            yield create_user.fill(
                username='context.username',
                email='context.email')
            if context.get('context.credit_card'):
                yield create_credit_card.fill(
                    username='context.username',
                    credit_card='context.credit_card')
            yield send_email.fill(email='context.email')
    """

    _decorate(fn, key='list', value=True)
    contextify(fn)
    return fn


def is_task_list(fn):
    """Check if a function is a task list.

    Return:
        boolean: if a function is a task list.
    """

    return getattr(fn, '__garcon__', {}).get('list')


def _decorate(fn, key=None, value=None):
    """Add the garcon property to the function.

    Args:
        fn (callable): The function to alter.
        key (string): The key to set (optional.)
        value (any): The value to set (optional.)
    """

    if not hasattr(fn, '__garcon__'):
        setattr(fn, '__garcon__', dict())

    if key:
        fn.__garcon__.update({
            key: value
        })


def _link_decorator(source_fn, dest_fn):
    """Link the garcon decorator values between two methods.

    If the destination method already have a value on `__garcon__`, we get it
    and merge it with the other one (so no values are lost.)

    Args:
        source_fn (callable): The method that contains `__garcon__`.
        dest_fn (callable): The method that receives the decorator.
    """

    source_values = copy.deepcopy(getattr(source_fn, '__garcon__', dict()))

    if hasattr(dest_fn, '__garcon__'):
        source_values.update(dest_fn.__garcon__)

    setattr(dest_fn, '__garcon__', source_values)


def contextify(fn):
    """Decorator to take values from the context and apply them to fn.

    The goal of this decorator is to allow methods to be called with different
    values from the same context. For instance: if you need to increase the
    throughtput of two different dynamodb tables, you will need to pass a
    table name, table index, and the new throughtput.

    If you have more than one table, it gets difficult to manage. With this
    decorator, it's a little easier::

        @contextify
        def increase_dynamodb_throughtput(
                activity, context, table_name=None, table_index=None,
                table_throughtput=None):
            print(table_name)
        activity_task = increase_dynamodb_throughtput.fill(
            table_name='dynamodb.table_name1',
            table_index='dynamodb.table_index1',
            table_throughtput='dynamodb.table_throughtput1')
        context = dict(
            'dynamodb.table_name1': 'table_name',
            'dynamodb.table_index1': 'index',
            'dynamodb.table_throughtput1': 'throughtput1')
        activity_task(..., context) # shows table_name
    """

    def fill(namespace=None, **requirements):

        requirements = {
            key: param.parametrize(current_param)
            for key, current_param in requirements.items()}

        def wrapper(context, **kwargs):
            kwargs.update(
                fill_function_call(
                    fn, requirements, kwargs.get('activity'), context))

            response = fn(**kwargs)
            if not response or not namespace:
                return response

            return namespace_result(response, namespace)

        update_wrapper(wrapper, fn)

        # Keep a record of the requirements value. This allows us to trim the
        # size of the context sent to the activity as an input.
        _link_decorator(fn, wrapper)
        _decorate(
            wrapper,
            'requirements',
            param.get_all_requirements(requirements.values()))
        return wrapper

    fn.fill = fill
    return fn


def flatten(callables, context=None):
    """Flatten the tasks.

    The task list is a mix of tasks and generators. The task generators are
    consuming the context and spawning new tasks. This method flattens
    everything into one list.

    Args:
        callables (list): list of callables (including tasks and generators.)

    Yield:
        callable: one of the task.
    """

    for task in callables:
        if is_task_list(task):
            for subtask in task(context):
                yield subtask
            continue
        yield task


def fill_function_call(fn, requirements, activity, context):
    """Fill a function calls from values from the context to the variable.

    Args:
        fn (callable): the function to call.
        requirements (dict): the requirements. The key represent the variable
            name and the value represents where the value is in the context.
        activity (ActivityWorker): the current activity worker.
        context (dict): the current context.

    Return:
        dict: The arguments to call the method with.
    """

    function_arguments = fn.__code__.co_varnames[:fn.__code__.co_argcount]
    kwargs = dict()

    for argument in function_arguments:
        param = requirements.get(argument, None)
        value = None

        if argument == 'context':
            raise Exception(
                'Data used from the context should be explicit. A task should'
                ' not randomly access information from the context.')

        elif argument == 'activity':
            value = activity

        elif param:
            value = param.get_data(context)

        kwargs.update({
            argument: value
        })

    return kwargs


def namespace_result(dictionary, namespace):
    """Namespace the response

    This method takes the keys in the map and add a prefix to all the keys
    (the namespace)::

        resp = dict(key='value', index='storage')
        namespace_response(resp, 'namespace')
        # Returns: {'namespace.index': 'storage', 'namespace.key': 'value'}

    Args:
        dictionary (dict): The dictionary to update.
        map (dict): The keys to update.

    Return:
        Dict: the updated dictionary
    """

    if not namespace:
        return dictionary

    return {
        (namespace + '.' + key): value for key, value in dictionary.items()}



from unittest.mock import MagicMock
from unittest.mock import ANY
import json
import sys

from botocore import exceptions
import pytest

from garcon import activity
from garcon import event
from garcon import runner
from garcon import task
from garcon import utils
from tests.fixtures import decider


def activity_run(
        monkeypatch, boto_client, poll=None, complete=None, fail=None,
        execute=None):
    """Create an activity.
    """

    current_activity = activity.Activity(boto_client)
    poll = poll or dict()

    monkeypatch.setattr(
        current_activity, 'execute_activity',
        execute or MagicMock(return_value=dict()))
    monkeypatch.setattr(
        boto_client, 'poll_for_activity_task', MagicMock(return_value=poll))
    monkeypatch.setattr(
        boto_client, 'respond_activity_task_completed', complete or MagicMock())
    monkeypatch.setattr(
        boto_client, 'respond_activity_task_failed', fail or MagicMock())

    return current_activity


@pytest.fixture(params=[0, 1, 2])
def generators(request):
    generators = []

    if request.param >= 1:
        def igenerator(context):
            for i in range(10):
                yield {'i': i}

        generators.append(igenerator)

    if request.param == 2:
        def dgenerator(context):
            for i in range(10):
                yield {'d': i * 2}
        generators.append(dgenerator)

    return generators


@pytest.fixture
def poll():
    return dict(
        activityId='something',
        taskToken='taskToken')


def test_poll_for_activity(monkeypatch, poll, boto_client):
    """Test that poll_for_activity successfully polls.
    """

    activity_task = poll
    current_activity = activity_run(monkeypatch, boto_client, poll)
    boto_client.poll_for_activity_task.return_value = activity_task

    activity_execution = current_activity.poll_for_activity()
    assert boto_client.poll_for_activity_task.called
    assert activity_execution.task_token is poll.get('taskToken')


def test_poll_for_activity_throttle_retry(monkeypatch, poll, boto_client):
    """Test that SWF throttles are retried during polling.
    """

    current_activity = activity_run(monkeypatch, boto_client, poll)
    boto_client.poll_for_activity_task.side_effect = exceptions.ClientError(
        {'Error': {'Code': 'ThrottlingException'}},
        'operation name')

    with pytest.raises(exceptions.ClientError):
        current_activity.poll_for_activity()
    assert boto_client.poll_for_activity_task.call_count == 5


def test_poll_for_activity_error(monkeypatch, poll, boto_client):
    """Test that non-throttle errors during poll are thrown.
    """

    current_activity = activity_run(monkeypatch, boto_client, poll)

    exception = Exception()
    boto_client.poll_for_activity_task.side_effect = exception

    with pytest.raises(Exception):
        current_activity.poll_for_activity()


def test_poll_for_activity_identity(monkeypatch, poll, boto_client):
    """Test that identity is passed to poll_for_activity.
    """

    current_activity = activity_run(monkeypatch, boto_client, poll)

    current_activity.poll_for_activity(identity='foo')
    boto_client.poll_for_activity_task.assert_called_with(
        domain=ANY, taskList=ANY, identity='foo')


def test_poll_for_activity_no_identity(monkeypatch, poll, boto_client):
    """Test poll_for_activity works without identity passed as param.
    """

    current_activity = activity_run(monkeypatch, boto_client, poll)

    current_activity.poll_for_activity()
    boto_client.poll_for_activity_task.assert_called_with(
        domain=ANY, taskList=ANY)


def test_run_activity(monkeypatch, poll, boto_client):
    """Run an activity.
    """

    current_activity = activity_run(monkeypatch, boto_client, poll=poll)
    current_activity.run()

    boto_client.poll_for_activity_task.assert_called_with(
        domain=ANY, taskList=ANY)
    assert current_activity.execute_activity.called
    assert boto_client.respond_activity_task_completed.called


def test_run_activity_identity(monkeypatch, poll, boto_client):
    """Run an activity with identity as param.
    """

    current_activity = activity_run(monkeypatch, boto_client, poll=poll)
    current_activity.run(identity='foo')

    boto_client.poll_for_activity_task.assert_called_with(
        domain=ANY, taskList=ANY, identity='foo')
    assert current_activity.execute_activity.called
    assert boto_client.respond_activity_task_completed.called


def test_run_capture_exception(monkeypatch, poll, boto_client):
    """Run an activity with an exception raised during activity execution.
    """

    current_activity = activity_run(monkeypatch, boto_client, poll)
    current_activity.on_exception = MagicMock()
    current_activity.execute_activity = MagicMock()
    error_msg_long = "Error" * 100
    actual_error_msg = error_msg_long[:255]
    current_activity.execute_activity.side_effect = Exception(error_msg_long)
    current_activity.run()

    assert boto_client.poll_for_activity_task.called
    assert current_activity.execute_activity.called
    assert current_activity.on_exception.called
    boto_client.respond_activity_task_failed.assert_called_with(
        taskToken=poll.get('taskToken'),
        reason=actual_error_msg)
    assert not boto_client.respond_activity_task_completed.called


def test_run_capture_fail_exception(monkeypatch, poll, boto_client):
    """Run an activity with an exception raised during failing execution.
    """

    current_activity = activity_run(monkeypatch, boto_client, poll)
    current_activity.on_exception = MagicMock()
    current_activity.execute_activity = MagicMock()
    current_activity.complete = MagicMock()
    current_activity.fail = MagicMock()
    error_msg_long = "Error" * 100
    current_activity.complete.side_effect = Exception(error_msg_long)
    current_activity.fail.side_effect = Exception(error_msg_long)
    current_activity.run()

    assert boto_client.poll_for_activity_task.called
    assert current_activity.execute_activity.called
    assert not current_activity.complete.called
    assert not current_activity.fail.called
    assert current_activity.on_exception.called


def test_run_capture_poll_exception(monkeypatch, boto_client, poll):
    """Run an activity with an exception raised during poll.
    """

    current_activity = activity_run(monkeypatch, boto_client, poll=poll)

    current_activity.on_exception = MagicMock()
    current_activity.execute_activity = MagicMock()

    exception = Exception('poll exception')
    boto_client.poll_for_activity_task.side_effect = exception
    current_activity.run()

    assert boto_client.poll_for_activity_task.called
    assert current_activity.on_exception.called
    assert not current_activity.execute_activity.called
    assert not boto_client.respond_activity_task_completed.called

    current_activity.on_exception = None
    current_activity.logger.error = MagicMock()
    current_activity.run()
    current_activity.logger.error.assert_called_with(exception, exc_info=True)


def test_run_activity_without_id(monkeypatch, boto_client):
    """Run an activity without an activity id.
    """

    current_activity = activity_run(monkeypatch, boto_client, poll=dict())
    current_activity.run()

    assert boto_client.poll_for_activity_task.called
    assert not current_activity.execute_activity.called
    assert not boto_client.respond_activity_task_completed.called


def test_run_activity_with_context(monkeypatch, boto_client, poll):
    """Run an activity with a context.
    """

    context = dict(foo='bar')
    poll.update(input=json.dumps(context))

    current_activity = activity_run(monkeypatch, boto_client, poll=poll)
    current_activity.run()

    activity_execution = current_activity.execute_activity.call_args[0][0]
    assert activity_execution.context == context


def test_run_activity_with_result(monkeypatch, boto_client, poll):
    """Run an activity with a result.
    """

    result = dict(foo='bar')
    mock = MagicMock(return_value=result)
    current_activity = activity_run(monkeypatch, boto_client, poll=poll,
        execute=mock)
    current_activity.run()
    boto_client.respond_activity_task_completed.assert_called_with(
        result=json.dumps(result), taskToken=poll.get('taskToken'))


def test_task_failure(monkeypatch, boto_client, poll):
    """Run an activity that has a bad task.
    """

    resp = dict(foo='bar')
    mock = MagicMock(return_value=resp)
    reason = 'fail'
    current_activity = activity_run(monkeypatch, boto_client, poll=poll,
        execute=mock)
    current_activity.on_exception = MagicMock()
    current_activity.execute_activity.side_effect = Exception(reason)
    current_activity.run()

    boto_client.respond_activity_task_failed.assert_called_with(
        taskToken=poll.get('taskToken'),
        reason=reason)


def test_task_failure_on_close_activity(monkeypatch, boto_client, poll):
    """Run an activity failure when the task is already closed.
    """

    resp = dict(foo='bar')
    mock = MagicMock(return_value=resp)
    current_activity = activity_run(monkeypatch, boto_client, poll=poll,
        execute=mock)
    current_activity.on_exception = MagicMock()
    current_activity.execute_activity.side_effect = Exception('fail')
    boto_client.respond_activity_task_failed.side_effect = Exception('fail')
    current_activity.unset_log_context = MagicMock()
    current_activity.run()

    assert current_activity.unset_log_context.called


def test_execute_activity(monkeypatch, boto_client):
    """Test the execution of an activity.
    """

    monkeypatch.setattr(activity.ActivityExecution, 'heartbeat',
        lambda self: None)

    resp = dict(task_resp='something')
    custom_task = MagicMock(return_value=resp)

    current_activity = activity.Activity(boto_client)
    current_activity.runner = runner.Sync(custom_task)

    val = current_activity.execute_activity(activity.ActivityExecution(
        boto_client, 'activityId', 'taskToken', '{"context": "value"}'))

    assert custom_task.called
    assert val == resp


def test_hydrate_activity(monkeypatch, boto_client):
    """Test the hydratation of an activity.
    """

    current_activity = activity.Activity(boto_client)
    current_activity.hydrate(dict(
        name='activity',
        domain='domain',
        requires=[],
        on_exception=lambda actor, exception: print(exception),
        tasks=[lambda: dict('val')]))


def test_create_activity(monkeypatch, boto_client):
    """Test the creation of an activity via `create`.
    """

    create = activity.create(boto_client, 'domain_name', 'flow_name')

    current_activity = create(name='activity_name')
    assert isinstance(current_activity, activity.Activity)
    assert current_activity.name == 'flow_name_activity_name'
    assert current_activity.task_list == 'flow_name_activity_name'
    assert current_activity.domain == 'domain_name'
    assert current_activity.client == boto_client


def test_create_external_activity(monkeypatch, boto_client):
    """Test the creation of an external activity via `create`.
    """

    create = activity.create(boto_client, 'domain_name', 'flow_name')

    current_activity = create(
        name='activity_name',
        timeout=60,
        heartbeat=40,
        external=True)

    assert isinstance(current_activity, activity.ExternalActivity)
    assert current_activity.name == 'flow_name_activity_name'
    assert current_activity.task_list == 'flow_name_activity_name'
    assert current_activity.domain == 'domain_name'

    assert isinstance(current_activity.runner, runner.External)
    assert current_activity.runner.heartbeat() == 40
    assert current_activity.runner.timeout() == 60


def test_create_activity_worker(monkeypatch):
    """Test the creation of an activity worker.
    """

    from tests.fixtures.flows import example

    worker = activity.ActivityWorker(example)
    assert len(worker.activities) == 4

    assert worker.flow is example
    assert not worker.worker_activities


def test_instances_creation(monkeypatch, boto_client, generators):
    """Test the creation of an activity instance id with the use of a local
    context.
    """

    local_activity = activity.Activity(boto_client)
    external_activity = activity.ExternalActivity(timeout=60)

    for current_activity in [local_activity, external_activity]:
        current_activity.generators = generators

        if len(current_activity.generators):
            instances = list(current_activity.instances(dict()))
            assert len(instances) == pow(10, len(generators))
            for instance in instances:
                assert isinstance(instance.local_context.get('i'), int)

                if len(generators) == 2:
                    assert isinstance(instance.local_context.get('d'), int)
        else:
            instances = list(current_activity.instances(dict()))
            assert len(instances) == 1
            assert isinstance(instances[0].local_context, dict)
            # Context is empty since no generator was used.
            assert not instances[0].local_context


def test_activity_timeouts(monkeypatch, boto_client, generators):
    """Test the creation of an activity timeouts.

    More details: the timeout of a task is 120s, the schedule to start is 1000,
    100 activities are going to be scheduled when the generator is set. The
    schedule_to_start for all activities instance is: 10000 * 100 = 100k. The
    schedule to close is 100k + duration of an activity (which is 120s * 2).
    """

    timeout = 120
    start_timeout = 1000

    @task.decorate(timeout=timeout)
    def local_task():
        return

    current_activity = activity.Activity(boto_client)
    current_activity.hydrate(dict(schedule_to_start=start_timeout))
    current_activity.generators = generators
    current_activity.runner = runner.Sync(
        local_task.fill(),
        local_task.fill())

    total_generators = pow(10, len(current_activity.generators))
    schedule_to_start = start_timeout * total_generators
    for instance in current_activity.instances({}):
        assert current_activity.pool_size == total_generators
        assert instance.schedule_to_start == schedule_to_start
        assert instance.timeout == timeout * 2
        assert instance.schedule_to_close == (
            schedule_to_start + instance.timeout)


def test_external_activity_timeouts(monkeypatch, boto_client, generators):
    """Test the creation of an external activity timeouts.
    """

    timeout = 120
    start_timeout = 1000

    current_activity = activity.ExternalActivity(timeout=timeout)
    current_activity.hydrate(dict(schedule_to_start=start_timeout))
    current_activity.generators = generators

    total_generators = pow(10, len(current_activity.generators))
    schedule_to_start = start_timeout * total_generators
    for instance in current_activity.instances({}):
        assert current_activity.pool_size == total_generators
        assert instance.schedule_to_start == schedule_to_start
        assert instance.timeout == timeout
        assert instance.schedule_to_close == (
            schedule_to_start + instance.timeout)


@pytest.mark.skipif(sys.version_info < (3, 0), reason="requires Python3")
def test_worker_run(monkeypatch, boto_client):
    """Test running the worker.
    """

    from tests.fixtures.flows import example

    worker = activity.ActivityWorker(example)
    assert len(worker.activities) == 4
    for current_activity in worker.activities:
        monkeypatch.setattr(
            current_activity, 'run', MagicMock(return_value=False))

    worker.run()

    assert len(worker.activities) == 4
    for current_activity in worker.activities:
        assert current_activity.run.called


def test_worker_run_with_skipped_activities(monkeypatch):
    """Test running the worker with defined activities.
    """

    monkeypatch.setattr(activity.Activity, 'run', MagicMock(return_value=False))

    from tests.fixtures.flows import example

    worker = activity.ActivityWorker(example, activities=['activity_1'])
    assert len(worker.worker_activities) == 1
    for current_activity in worker.activities:
        monkeypatch.setattr(
            current_activity, 'run', MagicMock(return_value=False))

    worker.run()

    for current_activity in worker.activities:
        if current_activity.name == 'activity_1':
            assert current_activity.run.called
        else:
            assert not current_activity.run.called


def test_worker_infinite_loop():
    """Test the worker runner.
    """

    spy = MagicMock()

    class Activity:
        def __init__(self):
            self.count = 0

        def run(self, identity=None):
            spy()
            self.count = self.count + 1
            if self.count < 5:
                return True
            return False

    activity_worker = Activity()
    activity_worker.name = 'activity_name'
    activity_worker.logger = MagicMock()
    activity.worker_runner(activity_worker)
    assert spy.called
    assert spy.call_count == 5


def test_worker_infinite_loop_on_external(monkeypatch):
    """There is no worker for external activities.
    """

    external_activity = activity.ExternalActivity(timeout=10)
    current_run = external_activity.run
    spy = MagicMock()

    def run():
        spy()
        return current_run()

    monkeypatch.setattr(external_activity, 'run', run)
    activity.worker_runner(external_activity)

    # This test might not fail, but it will hang the test suite since it is
    # going to trigger an infinite loop.
    assert spy.call_count == 1


def test_activity_launch_sequence():
    """Test available activities.
    """

    from tests.fixtures.flows import example

    # First available activity is the activity_1.
    context = dict()
    history = event.activity_states_from_events(decider.history['events'][:1])
    activities = list(
        activity.find_available_activities(example, history, context))
    uncomplete = list(
        activity.find_uncomplete_activities(example, history, context))
    assert len(activities) == 1
    assert len(uncomplete) == 4
    assert activities[0].activity_worker == example.activity_1

    # In between activities should not launch activities.
    history = event.activity_states_from_events(decider.history['events'][:5])
    activities = list(
        activity.find_available_activities(example, history, context))
    uncomplete = list(
        activity.find_uncomplete_activities(example, history, context))
    assert len(activities) == 0
    assert len(uncomplete) == 4

    # Two activities are launched in parallel: 2 and 3.
    history = event.activity_states_from_events(decider.history['events'][:7])
    activities = list(
        activity.find_available_activities(example, history, context))
    uncomplete = list(
        activity.find_uncomplete_activities(example, history, context))
    assert len(activities) == 2
    assert example.activity_1 not in uncomplete

    # Activity 3 completes before activity 2. Activity 4 depends on 2 and 3 to
    # complete.
    history = event.activity_states_from_events(decider.history['events'][:14])
    activities = list(
        activity.find_available_activities(example, history, context))
    uncomplete = list(
        activity.find_uncomplete_activities(example, history, context))
    assert len(activities) == 0
    assert example.activity_3 not in uncomplete

    # Activity 2 - 3 completed.
    history = event.activity_states_from_events(decider.history['events'][:22])
    activities = list(
        activity.find_available_activities(example, history, context))
    uncomplete = list(
        activity.find_uncomplete_activities(example, history, context))
    assert len(activities) == 1
    assert activities[0].activity_worker == example.activity_4
    assert example.activity_1 not in uncomplete
    assert example.activity_2 not in uncomplete
    assert example.activity_3 not in uncomplete

    # Close
    history = event.activity_states_from_events(decider.history['events'][:25])
    activities = list(
        activity.find_available_activities(example, history, context))
    uncomplete = list(
        activity.find_uncomplete_activities(example, history, context))
    assert not activities
    assert not uncomplete


def test_create_activity_instance():
    """Test the creation of an activity instance.
    """

    activity_mock = MagicMock()
    activity_mock.name = 'foobar'
    activity_mock.retry = 20

    instance = activity.ActivityInstance(activity_mock)

    assert activity_mock.name == instance.activity_name
    assert activity_mock.retry == instance.retry


def test_create_activity_instance_id(monkeypatch):
    """Test the creation of an activity instance id.
    """

    monkeypatch.setattr(utils, 'create_dictionary_key', MagicMock())

    activity_mock = MagicMock()
    activity_mock.name = 'activity'
    instance = activity.ActivityInstance(activity_mock)

    # No context was passed, so create_dictionary key didn't need to be
    # called.
    assert instance.id == activity_mock.name + '-1'
    assert not utils.create_dictionary_key.called


def test_create_activity_instance_id_with_local_context(monkeypatch):
    """Test the creation of an activity instance id with the use of a local
    context.
    """

    monkeypatch.setattr(utils, 'create_dictionary_key', MagicMock())

    activity_mock = MagicMock()
    activity_mock.name = 'activity'
    instance = activity.ActivityInstance(activity_mock, dict(foobar='yes'))

    assert instance.id.startswith(activity_mock.name)
    assert utils.create_dictionary_key.called


def test_create_activity_instance_input_without_runner(monkeypatch):
    """Test the creation of a context for an activity instance input without
    specifying a runner.
    """

    activity_mock = MagicMock()
    activity_mock.name = 'activity'
    activity_mock.runner = None
    context = dict(context='yes')
    instance = activity.ActivityInstance(activity_mock, context)

    with pytest.raises(runner.RunnerMissing):
        instance.create_execution_input()


def test_create_activity_instance_input(monkeypatch):
    """Test the creation of a context for an activity instance input.
    """

    @task.decorate()
    def task_a(value):
        pass

    activity_mock = MagicMock()
    activity_mock.name = 'activity'
    activity_mock.runner = runner.BaseRunner(task_a.fill(value='context'))
    instance = activity.ActivityInstance(
        activity_mock, local_context=dict(context='yes', unused='no'),
        execution_context=dict(somemore='values'))
    resp = instance.create_execution_input()

    assert len(resp) == 4
    assert resp.get('context') == 'yes'
    assert 'somemore' not in resp
    assert 'unused' not in resp
    assert 'execution.domain' in resp
    assert 'execution.run_id' in resp
    assert 'execution.workflow_id' in resp


def test_create_activity_instance_input_without_decorate(monkeypatch):
    """Test the creation of a context input without the use of a decorator.
    """

    def task_a(value):
        pass

    activity_mock = MagicMock()
    activity_mock.name = 'activity'
    context = dict(foo='bar')
    local_context = dict(context='yes')

    activity_mock.runner = runner.BaseRunner(task_a)
    instance = activity.ActivityInstance(
        activity_mock, local_context=local_context,
        execution_context=context)

    resp = instance.create_execution_input()
    assert resp.get('foo') == 'bar'
    assert resp.get('context') == 'yes'


def test_create_activity_instance_input_with_zero_or_empty_values(
        monkeypatch):
    """Test the creation of a context for an activity instance input.
    """

    @task.decorate()
    def task_a(value1, value2, value3, value4):
        pass

    activity_mock = MagicMock()
    activity_mock.name = 'activity'
    activity_mock.runner = runner.BaseRunner(
        task_a.fill(
            value1='zero',
            value2='empty_list',
            value3='empty_dict',
            value4='none'))
    instance = activity.ActivityInstance(
        activity_mock,
        local_context=dict(
            zero=0, empty_list=[], empty_dict={}, none=None))

    resp = instance.create_execution_input()

    assert len(resp) == 6
    assert resp.get('zero') == 0
    assert resp.get('empty_list') == []
    assert resp.get('empty_dict') == {}
    assert 'none' not in resp


def test_activity_state():
    """Test the creation of the activity state.
    """

    activity_id = 'id'
    state = activity.ActivityState(activity_id)
    assert state.activity_id is activity_id
    assert not state.get_last_state()

    state.add_state(activity.ACTIVITY_FAILED)
    state.add_state(activity.ACTIVITY_COMPLETED)
    assert len(state.states) == 2
    assert state.get_last_state() is activity.ACTIVITY_COMPLETED

    result = 'foobar'
    state.set_result(result)
    assert state.result == result

    with pytest.raises(Exception):
        state.set_result('shouldnt reset')

    assert state.result == result



from unittest.mock import MagicMock

import pytest

from garcon import activity
from garcon import runner
from garcon import task


EMPTY_CONTEXT = dict()


def test_execute_default_task_runner():
    """Should throw an exception.
    """

    current_runner = runner.BaseRunner()
    with pytest.raises(NotImplementedError):
        current_runner.execute(None, None)


def test_synchronous_tasks(monkeypatch, boto_client):
    """Test synchronous tasks.
    """

    monkeypatch.setattr(activity.ActivityExecution, 'heartbeat',
        lambda self: None)

    resp = dict(foo='bar')
    current_runner = runner.Sync(
        MagicMock(), MagicMock(return_value=resp))
    current_activity = activity.ActivityExecution(
        boto_client, 'activityId', 'taskToken', '{"context": "value"}')

    result = current_runner.execute(current_activity, EMPTY_CONTEXT)

    assert len(current_runner.tasks) == 2

    for current_task in task.flatten(current_runner.tasks, EMPTY_CONTEXT):
        assert current_task.called

    assert resp == result


def test_aynchronous_tasks(monkeypatch, boto_client):
    """Test asynchronous tasks.
    """

    monkeypatch.setattr(activity.ActivityExecution, 'heartbeat',
        lambda self: None)

    tasks = [MagicMock() for i in range(5)]
    tasks[2].return_value = dict(oi='mondo')
    tasks[4].return_value = dict(bonjour='monde')
    expected_response = dict(
        list(tasks[2].return_value.items()) +
        list(tasks[4].return_value.items()))

    workers = 2
    current_runner = runner.Async(*tasks, max_workers=workers)

    assert current_runner.max_workers == workers
    assert len(current_runner.tasks) == len(tasks)

    current_activity = activity.ActivityExecution(boto_client,
         'activityId', 'taskToken', '{"hello": "world"}')

    resp = current_runner.execute(current_activity, current_activity.context)

    for current_task in tasks:
        assert current_task.called

    assert resp == expected_response


def test_calculate_timeout_with_no_tasks():
    """Task list without task has no timeout.
    """

    task_list = runner.BaseRunner()
    assert task_list.timeout(EMPTY_CONTEXT) == 0


def test_calculate_heartbeat_with_no_tasks():
    """Task list without tasks has no heartbeat.
    """

    task_list = runner.BaseRunner()
    assert task_list.heartbeat(EMPTY_CONTEXT) == 0


def test_calculate_default_timeout():
    """Tasks that do not have a set timeout get the default timeout.
    """

    task_list = runner.BaseRunner(lambda x: x)
    assert task_list.timeout(EMPTY_CONTEXT) == runner.DEFAULT_TASK_TIMEOUT


def test_calculate_default_heartbeat():
    """Tasks that do not have a set timeout get the default timeout.
    """

    task_list = runner.BaseRunner(lambda x: x)
    assert task_list.heartbeat(EMPTY_CONTEXT) == runner.DEFAULT_TASK_HEARTBEAT


def test_calculate_timeout():
    """Check methods that have set timeout.
    """

    timeout = 10

    @task.timeout(timeout)
    def task_a():
        pass

    current_runner = runner.BaseRunner(task_a)
    assert current_runner.timeout(EMPTY_CONTEXT) == timeout

    @task.decorate(timeout=timeout)
    def task_b():
        pass

    current_runner = runner.BaseRunner(task_b)
    assert current_runner.timeout(EMPTY_CONTEXT) == timeout

    def task_c():
        pass

    current_runner = runner.BaseRunner(task_a, task_c)
    current_timeout = current_runner.timeout(EMPTY_CONTEXT)
    expected_timeout = timeout + runner.DEFAULT_TASK_TIMEOUT
    assert current_timeout == expected_timeout


def test_calculate_heartbeat():
    """Check methods that have set timeout.
    """

    @task.decorate(heartbeat=5)
    def task_a():
        pass

    current_runner = runner.BaseRunner(task_a)
    assert current_runner.heartbeat(EMPTY_CONTEXT) == 5

    @task.decorate(heartbeat=3)
    def task_b():
        pass

    current_runner = runner.BaseRunner(task_b)
    assert current_runner.heartbeat(EMPTY_CONTEXT) == 3

    @task.decorate(heartbeat=4498)
    def task_c():
        pass

    def task_d():
        pass

    current_runner = runner.BaseRunner(
        task_a, task_b, task_c, task_d)
    assert current_runner.heartbeat(EMPTY_CONTEXT) == 4498


def test_runner_requirements():
    """Test the requirements for the runner
    """

    @task.decorate()
    def task_a():
        pass

    @task.decorate(timeout=20)
    def task_b():
        pass

    value_1 = 'context.value'
    value_2 = 'context.value_1'
    current_runner = runner.BaseRunner(
        task_a.fill(foo=value_1),
        task_b.fill(foobar=value_2))

    requirements = current_runner.requirements(EMPTY_CONTEXT)
    assert len(requirements) == 2
    assert value_1 in requirements
    assert value_2 in requirements


def test_runner_requirements_without_decoration():
    """Should just throw an exception.
    """

    def task_a():
        pass

    current_runner = runner.BaseRunner(task_a)

    with pytest.raises(runner.NoRunnerRequirementsFound):
        current_runner.requirements(EMPTY_CONTEXT)



from setuptools import find_packages
from setuptools import setup
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))
# Get the long description from the README file
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()


setup(
    name='Garcon',
    version='1.0.3',
    url='https://github.com/xethorn/garcon/',
    author='Michael Ortali',
    author_email='hello@xethorn.net',
    description=('Lightweight library for AWS SWF.'),
    long_description=long_description,
    license='MIT',
    packages=find_packages(),
    include_package_data=True,
    install_requires=['boto3', 'backoff'],
    zip_safe=False,
    classifiers=[
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
    ],)



from garcon import activity
from garcon import runner

import boto3

client = boto3.client('swf', region_name='us-east-1')
domain = 'dev'
name = 'workflow_name'
create = activity.create(client, domain, name)

activity_1 = create(
    name='activity_1',
    tasks=runner.Sync(
        lambda activity, context:
            print('activity_1')))

activity_2 = create(
    name='activity_2',
    requires=[activity_1],
    tasks=runner.Async(
        lambda activity, context:
            print('activity_2_task_1'),
        lambda activity, context:
            print('activity_2_task_2')))

activity_3 = create(
    name='activity_3',
    requires=[activity_1],
    tasks=runner.Sync(
        lambda activity, context:
            print('activity_3')))

activity_4 = create(
    name='activity_4',
    requires=[activity_3, activity_2],
    tasks=runner.Sync(
        lambda activity, context:
            print('activity_4')))


def on_exception(actor, exception):
    """Handler for exceptions.

    Useful if you use sentry or other similar systems.
    """

    print(exception)



from unittest.mock import MagicMock
import json

import pytest

from garcon import decider
from garcon import activity
from tests.fixtures import decider as decider_events


def test_create_decider(monkeypatch):
    """Create a decider and check the behavior of the registration.
    """

    from tests.fixtures.flows import example

    d = decider.DeciderWorker(example)
    assert d.client
    assert len(d.activities) == 4
    assert d.flow
    assert d.domain
    assert d.on_exception

    monkeypatch.setattr(decider.DeciderWorker, 'register', MagicMock())
    d = decider.DeciderWorker(example)
    assert d.register.called

    monkeypatch.setattr(decider.DeciderWorker, 'register', MagicMock())
    dec = decider.DeciderWorker(example, register=False)
    assert not dec.register.called


def test_get_history(monkeypatch):
    """Test the decider history
    """

    from tests.fixtures.flows import example

    events = decider_events.history.get('events')
    half = int(len(events) / 2)
    events = events[:half * 2]
    pool_1 = events[half:]
    pool_2 = events[:half]
    identity = 'identity'

    d = decider.DeciderWorker(example)
    d.client.poll_for_decision_task = MagicMock(return_value={'events': pool_2})

    resp = d.get_history(
        identity, {'events': pool_1, 'nextPageToken': 'nextPage'})

    d.client.poll_for_decision_task.assert_called_with(
        domain=example.domain,
        nextPageToken='nextPage',
        identity=identity,
        taskList=dict(name=d.task_list))
    assert len(resp) == len([
        evt for evt in events if evt['eventType'].startswith('Decision')])


def test_get_activity_states(monkeypatch):
    """Test get activity states from history.
    """

    from tests.fixtures.flows import example

    identity= 'identity'
    events = decider_events.history.get('events')
    d = decider.DeciderWorker(example)
    history = d.get_history(identity, {'events': events})
    states = d.get_activity_states(history)

    for activity_name, activity_instances in states.items():
        for activity_instance, activity_state in activity_instances.items():
            assert isinstance(activity_state, activity.ActivityState)


def test_running_workflow(monkeypatch):
    """Test running a workflow.
    """

    from tests.fixtures.flows import example

    d = decider.DeciderWorker(example, register=False)
    d.client.poll_for_decision_task = MagicMock(
        return_value=decider_events.history)
    d.client.respond_decision_task_completed = MagicMock()
    d.complete = MagicMock()
    d.create_decisions_from_flow = MagicMock()

    # Via flow.
    d.run()
    assert d.create_decisions_from_flow.called

    # Via custom decider
    spy = MagicMock()

    def custom_decider(schedule):
        spy()

    example.decider = custom_decider
    d.run()
    assert spy.called


def test_running_workflow_identity(monkeypatch):
    """Test running a workflow with and without identity.
    """

    from tests.fixtures.flows import example

    d = decider.DeciderWorker(example, register=False)
    d.client.poll_for_decision_task = MagicMock()
    d.complete = MagicMock()
    d.create_decisions_from_flow = MagicMock()

    # assert running decider without identity
    d.run()
    d.client.poll_for_decision_task.assert_called_with(
        domain=d.domain,
        taskList=dict(name=d.task_list),
        identity='')

    # assert running decider with identity
    d.run('foo')
    d.client.poll_for_decision_task.assert_called_with(
        domain=d.domain,
        taskList=dict(name=d.task_list),
        identity='foo')


def test_running_workflow_exception(monkeypatch):
    """Run a decider with an exception raised during poll.
    """

    from tests.fixtures.flows import example

    d = decider.DeciderWorker(example, register=False)
    d.client.poll_for_decision_task = MagicMock(
        return_value=decider_events.history)
    d.complete = MagicMock()
    d.create_decisions_from_flow = MagicMock()
    exception = Exception('test')
    d.client.poll_for_decision_task.side_effect = exception
    d.on_exception = MagicMock()
    d.logger.error = MagicMock()
    d.run()
    assert d.on_exception.called
    d.logger.error.assert_called_with(exception, exc_info=True)
    assert not d.complete.called


def test_create_decisions_from_flow_exception(monkeypatch):
    """Test exception is raised and workflow fails when exception raised.
    """

    from tests.fixtures.flows import example

    decider_worker = decider.DeciderWorker(example, register=False)
    decider_worker.logger.error = MagicMock()
    decider_worker.on_exception = MagicMock()

    exception = Exception('test')
    monkeypatch.setattr(decider.activity,
        'find_available_activities', MagicMock(side_effect = exception))

    decisions = []
    mock_activity_states = MagicMock()
    mock_context = MagicMock()
    decider_worker.create_decisions_from_flow(
        decisions, mock_activity_states, mock_context)

    failed_execution = dict(
        decisionType='FailWorkflowExecution',
        failWorkflowExecutionDecisionAttributes=dict(reason=str(exception)))

    assert failed_execution in decisions
    assert decider_worker.on_exception.called
    decider_worker.logger.error.assert_called_with(exception, exc_info=True)


def test_running_workflow_without_events(monkeypatch):
    """Test running a workflow without having any events.
    """

    from tests.fixtures.flows import example

    d = decider.DeciderWorker(example, register=False)
    d.client.poll_for_decision_task = MagicMock(return_value={})
    d.get_history = MagicMock()
    d.run()

    assert not d.get_history.called


def test_schedule_context():
    """Test the schedule context.
    """

    context = decider.ScheduleContext()
    assert context.completed

    context.mark_uncompleted()
    assert not context.completed


def test_schedule_with_unscheduled_activity(monkeypatch):
    """Test the scheduling of an activity.
    """

    from tests.fixtures.flows import example

    monkeypatch.setattr(decider, 'schedule_activity_task', MagicMock())

    decisions = MagicMock()
    schedule_context = decider.ScheduleContext()
    history = {}
    current_activity = example.activity_1

    decider.schedule(
        decisions, schedule_context, history, {}, 'schedule_id',
        current_activity)

    assert decider.schedule_activity_task.called
    assert not schedule_context.completed


def test_schedule_with_scheduled_activity(monkeypatch):
    """Test the scheduling of an activity.
    """

    from tests.fixtures.flows import example

    monkeypatch.setattr(decider, 'schedule_activity_task', MagicMock())

    decisions = MagicMock()
    schedule_context = decider.ScheduleContext()
    instance_state = activity.ActivityState('activity_1')
    instance_state.add_state(activity.ACTIVITY_SCHEDULED)
    current_activity = example.activity_1
    history = {
        current_activity.name: {
            'workflow_name_activity_1-1-schedule_id': instance_state
        }
    }

    resp = decider.schedule(
        decisions, schedule_context, history, {}, 'schedule_id',
        current_activity)

    assert not decider.schedule_activity_task.called
    assert not schedule_context.completed
    assert resp.get_last_state() == activity.ACTIVITY_SCHEDULED

    with pytest.raises(activity.ActivityInstanceNotReadyException):
        resp.result.get('foo')


def test_schedule_with_completed_activity(monkeypatch):
    """Test the scheduling of an activity.
    """

    from tests.fixtures.flows import example

    monkeypatch.setattr(decider, 'schedule_activity_task', MagicMock())

    decisions = MagicMock()
    schedule_context = decider.ScheduleContext()
    instance_state = activity.ActivityState('activity_1')
    instance_state.add_state(activity.ACTIVITY_COMPLETED)
    current_activity = example.activity_1
    history = {
        current_activity.name: {
            'workflow_name_activity_1-1-schedule_id': instance_state
        }
    }

    resp = decider.schedule(
        decisions, schedule_context, history, {}, 'schedule_id',
        current_activity)

    assert not decider.schedule_activity_task.called
    assert resp.get_last_state() == activity.ACTIVITY_COMPLETED
    assert schedule_context.completed
    resp.result.get('foo')


def test_schedule_requires_with_incomplete_activities():
    """Test the scheduler.
    """

    activity_state = activity.ActivityState('activity_name')
    with pytest.raises(activity.ActivityInstanceNotReadyException):
        decider.ensure_requirements([activity_state])

    with pytest.raises(activity.ActivityInstanceNotReadyException):
        decider.ensure_requirements([None])

    activity_state.add_state(activity.ACTIVITY_COMPLETED)
    decider.ensure_requirements(requires=[activity_state])


def test_schedule_activity_task(monkeypatch):
    """Test scheduling an activity task.
    """

    from tests.fixtures.flows import example

    instance = list(example.activity_1.instances({}))[0]
    decisions = []
    decider.schedule_activity_task(decisions, instance)
    expects = dict(
        decisionType='ScheduleActivityTask',
        scheduleActivityTaskDecisionAttributes=dict(
            activityId=instance.id,
            activityType=dict(
                name=instance.activity_name,
                version='1.0'),
            taskList=dict(name=instance.activity_worker.task_list),
            input=json.dumps(instance.create_execution_input()),
            heartbeatTimeout=str(instance.heartbeat_timeout),
            startToCloseTimeout=str(instance.timeout),
            scheduleToStartTimeout=str(instance.schedule_to_start),
            scheduleToCloseTimeout=str(instance.schedule_to_close)))
    assert expects in decisions


def test_schedule_activity_task_with_version(monkeypatch):
    """Test scheduling an activity task with a version.
    """

    from tests.fixtures.flows import example

    instance = list(example.activity_1.instances({}))[0]
    decisions = []
    version = '2.0'
    decider.schedule_activity_task(decisions, instance, version=version)
    expects = dict(
        decisionType='ScheduleActivityTask',
        scheduleActivityTaskDecisionAttributes=dict(
            activityId=instance.id,
            activityType=dict(
                name=instance.activity_name,
                version=version),
            taskList=dict(name=instance.activity_worker.task_list),
            input=json.dumps(instance.create_execution_input()),
            heartbeatTimeout=str(instance.heartbeat_timeout),
            startToCloseTimeout=str(instance.timeout),
            scheduleToStartTimeout=str(instance.schedule_to_start),
            scheduleToCloseTimeout=str(instance.schedule_to_close)))
    assert expects in decisions


def test_schedule_activity_task_with_custom_id(monkeypatch):
    """Test scheduling an activity task with a custom id.
    """

    from tests.fixtures.flows import example

    instance = list(example.activity_1.instances({}))[0]
    decisions = []
    custom_id = 'special_id'
    decider.schedule_activity_task(decisions, instance, id=custom_id)
    expects = dict(
        decisionType='ScheduleActivityTask',
        scheduleActivityTaskDecisionAttributes=dict(
            activityId=custom_id,
            activityType=dict(
                name=instance.activity_name,
                version='1.0'),
            taskList=dict(name=instance.activity_worker.task_list),
            input=json.dumps(instance.create_execution_input()),
            heartbeatTimeout=str(instance.heartbeat_timeout),
            startToCloseTimeout=str(instance.timeout),
            scheduleToStartTimeout=str(instance.schedule_to_start),
            scheduleToCloseTimeout=str(instance.schedule_to_close)))
    assert expects in decisions



from garcon import log


class MockLogClient(log.GarconLogger):
    """Mock of an object for which we want to add a Garcon logger
    """
    domain = 'test_domain'


# Valid execution context
execution_context = {
    'execution.domain': 'dev',
    'execution.run_id': '123abc=',
    'execution.workflow_id': 'test-workflow-id'}

# Invalid execution context. Keys are incorrect
invalid_execution_context = {
    'abcd.domain': 'dev',
    '123.run_id': '123abc=',
    'XYZ.workflow_id': 'test-workflow-id'}


def log_enabled_object():
    """Creates a mock object with log enabled
    """

    mock = MockLogClient()
    mock.set_log_context(execution_context)

    return mock


def log_disabled_object():
    """Creates a mock object with no log
    """

    mock = MockLogClient()
    mock.set_log_context(invalid_execution_context)

    return mock
